import json

import os
import traceback
import asyncio

from typing import Any, Dict, List, Tuple, Union

from mcp.types import TextContent, ImageContent

from aworld.core.common import ActionModel, ActionResult, Observation
from aworld.core.tool.base import ToolActionExecutor, Tool, AsyncTool
from aworld.logs.util import logger
from aworld.mcp_client.server import MCPServer, MCPServerSse
import aworld.mcp_client.utils as mcp_utils
from aworld.utils.common import sync_exec, find_file


class MCPToolExecutor(ToolActionExecutor):
    """A tool executor that uses MCP server to execute actions."""

    def __init__(self, tool: Union[Tool, AsyncTool] = None):
        """Initialize the MCP tool executor."""
        super().__init__(tool)
        self.initialized = False
        self.mcp_servers: Dict[str, MCPServer] = {}
        self._load_mcp_config()

    def _replace_env_variables(self, config):
        if isinstance(config, dict):
            for key, value in config.items():
                if isinstance(value, str) and value.startswith("${") and value.endswith("}"):
                    env_var_name = value[2:-1]
                    config[key] = os.getenv(env_var_name, value)
                    logger.info(f"Replaced {value} with {config[key]}")
                elif isinstance(value, dict) or isinstance(value, list):
                    self._replace_env_variables(value)
        elif isinstance(config, list):
            for index, item in enumerate(config):
                if isinstance(item, str) and item.startswith("${") and item.endswith("}"):
                    env_var_name = item[2:-1]
                    config[index] = os.getenv(env_var_name, item)
                    logger.info(f"Replaced {item} with {config[index]}")
                elif isinstance(item, dict) or isinstance(item, list):
                    self._replace_env_variables(item)

    def _load_mcp_config(self) -> None:
        """Load MCP server configurations from config file."""
        try:
            config_data = {}
            if mcp_utils.MCP_SERVERS_CONFIG:
                config_data=mcp_utils.MCP_SERVERS_CONFIG
            else:
                # Priority given to the running path.
                config_path = find_file(filename='mcp.json')
                if not os.path.exists(config_path):
                    # Use relative path for config file
                    current_dir = os.path.dirname(os.path.abspath(__file__))
                    config_path = os.path.normpath(os.path.join(current_dir, "../../config/mcp.json"))
                logger.info(f"mcp conf path: {config_path}")

                with open(config_path, "r") as f:
                    config_data = json.load(f)

                # Replace environment variables in the configuration
                self._replace_env_variables(config_data)

            # Load all server configurations
            for server_name, server_config in config_data.get("mcpServers", {}).items():
                # Skip disabled servers
                if server_config.get("disabled", False):
                    continue

                # Handle SSE server
                if "url" in server_config:
                    self.mcp_servers[server_name] = {
                        "type": "sse",
                        "url": server_config["url"],
                        "instance": None,
                        "timeout": server_config.get('timeout', 5.),
                        "sse_read_timeout": server_config.get('sse_read_timeout', 300.0),
                        "headers": server_config.get('headers')
                    }
                # Handle stdio server
                elif "command" in server_config:
                    self.mcp_servers[server_name] = {
                        "type": "stdio",
                        "command": server_config["command"],
                        "args": server_config.get("args", []),
                        "env": server_config.get("env", {}),
                        "cwd": server_config.get("cwd"),
                        "encoding": server_config.get("encoding", "utf-8"),
                        "encoding_error_handler": server_config.get("encoding_error_handler", "strict"),
                        "instance": None
                    }

            self.initialized = True
        except Exception as e:
            logger.error(f"Failed to load MCP config: {traceback.format_exc()}")

    async def _get_or_create_server(self, server_name: str) -> MCPServer:
        """Get an existing MCP server instance or create a new one."""
        if server_name not in self.mcp_servers:
            raise ValueError(f"MCP server '{server_name}' not found in configuration")

        server_info = self.mcp_servers[server_name]

        # If an instance already exists, check if it's available and reuse it
        if server_info.get("instance"):
            return server_info["instance"]

        server_type = server_info.get("type", "sse")

        try:
            if server_type == "sse":
                # Create new SSE server instance
                server_params = {
                    "url": server_info["url"],
                    "timeout": server_info['timeout'],
                    "sse_read_timeout": server_info['sse_read_timeout'],
                    "headers": server_info['headers']
                }

                server = MCPServerSse(server_params, cache_tools_list=True, name=server_name)
            elif server_type == "stdio":
                # Create new stdio server instance
                server_params = {
                    "command": server_info["command"],
                    "args": server_info["args"],
                    "env": server_info["env"],
                    "cwd": server_info.get("cwd"),
                    "encoding": server_info["encoding"],
                    "encoding_error_handler": server_info["encoding_error_handler"]
                }

                from aworld.mcp_client.server import MCPServerStdio
                server = MCPServerStdio(server_params, cache_tools_list=True, name=server_name)
            else:
                raise ValueError(f"Unsupported MCP server type: {server_type}")

            # Try to connect, with special handling for cancellation exceptions
            try:
                await server.connect()
            except asyncio.CancelledError:
                # When the task is cancelled, ensure resources are cleaned up
                logger.warning(f"Connection to server '{server_name}' was cancelled")
                await server.cleanup()
                raise

            server_info["instance"] = server
            return server

        except asyncio.CancelledError:
            # Pass cancellation exceptions up to be handled by the caller
            raise
        except Exception as e:
            logger.error(f"Failed to connect to MCP server '{server_name}': {e}")
            raise

    async def async_execute_action(self, actions: List[ActionModel], **kwargs) -> Tuple[
        List[ActionResult], Any]:
        """Execute actions using the MCP server.

        Args:
            actions: A list of action models to execute
            **kwargs: Additional arguments

        Returns:
            A list of action results
        """
        if not self.initialized:
            raise RuntimeError("MCP Tool Executor not initialized")

        if not actions:
            return [], None

        results = []
        for action in actions:
            # Get server and operation information
            server_name = action.tool_name
            if not server_name:
                raise ValueError("Missing tool_name in action model")

            action_name = action.action_name
            if not action_name:
                raise ValueError("Missing action_name in action model")

            params = action.params or {}

            try:
                server = self.mcp_servers.get(server_name, {}).get('instance', None)
                if not server:
                    # Get or create MCP server
                    server = await self._get_or_create_server(server_name)

                # Call the tool and process results
                try:
                    result = await server.call_tool(action_name, params)

                    if result and result.content:
                        if isinstance(result.content[0], TextContent):
                            action_result = ActionResult(
                                content=result.content[0].text,
                                keep=True
                            )
                        elif isinstance(result.content[0], ImageContent):
                            action_result = ActionResult(
                                content=f"data:image/jpeg;base64,{result.content[0].data}",
                                keep=True
                            )
                        else:
                            action_result = ActionResult(
                                content="",
                                keep=True
                            )
                            logger.warning("Unsupported content type is error:")
                    else:
                        action_result = ActionResult(
                            content="",
                            keep=True
                        )
                        logger.warning("mcp result is null")

                    results.append(action_result)
                except asyncio.CancelledError:
                    # Log cancellation exception, reset server connection to avoid async context confusion
                    logger.warning(f"Tool call to {action_name} on {server_name} was cancelled")
                    if server_name in self.mcp_servers and self.mcp_servers[server_name].get("instance"):
                        try:
                            await self.mcp_servers[server_name]["instance"].cleanup()
                            self.mcp_servers[server_name]["instance"] = None
                        except Exception as cleanup_error:
                            logger.error(f"Error cleaning up server after cancellation: {cleanup_error}")
                    # Re-raise exception to notify upper level caller
                    raise

            except asyncio.CancelledError:
                # Pass cancellation exception
                logger.warning("Async execution was cancelled")
                raise

            except Exception as e:
                # Handle general errors
                error_msg = str(e)
                logger.error(f"Error executing MCP action: {error_msg}")
                action_result = ActionResult(
                    content=f"Error executing tool: {error_msg}",
                    keep=True
                )
                results.append(action_result)

        return results, None

    async def cleanup(self) -> None:
        """Clean up all MCP server connections."""
        for server_name, server_info in self.mcp_servers.items():
            if server_info.get("instance"):
                try:
                    await server_info["instance"].cleanup()
                except Exception as e:
                    logger.error(f"Error cleaning up MCP server {server_name}: {e}")

    async def close(self, keys: List[str] = []) -> None:
        if keys:
            for key in keys:
                if key in self.mcp_servers:
                    server_info = self.mcp_servers[key]
                    # Fixme: Have resources leak, MCP server may clean fail.
                    if server_info.get("type") == "stdio":
                        server_info["instance"] = None
                        continue

                    try:
                        await server_info["instance"].cleanup()
                    except Exception as e:
                        logger.error(f"Error cleaning up MCP server {key}: {e}")

    def execute_action(self, actions: List[ActionModel], **kwargs) -> Tuple[
        List[ActionResult], Any]:
        return sync_exec(self.async_execute_action, actions, **kwargs)



# coding: utf-8
# Copyright (c) 2025 inclusionAI.
import abc

from typing import TypeVar, Generic, AsyncGenerator

from aworld.core.event.base import Message

IN = TypeVar('IN')
OUT = TypeVar('OUT')


class Handler(Generic[IN, OUT]):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    async def handle(self, data: IN) -> AsyncGenerator[OUT, None]:
        """Process the data as the expected result.

        Args:
            data: Data generated while running the task.
        """

    @classmethod
    def name(cls):
        """Handler name."""
        return cls.__name__


class DefaultHandler(Handler[Message, AsyncGenerator[Message, None]]):
    """Default handler."""



import base64
import json
import os
import traceback
from typing import List

from mcp.server.fastmcp import FastMCP
from openai import OpenAI
from pydantic import Field

from aworld.logs.util import logger
from mcp_servers.utils import get_file_from_source

# Initialize MCP server
mcp = FastMCP("audio-server")


client = OpenAI(
    api_key=os.getenv("AUDIO_LLM_API_KEY"), base_url=os.getenv("AUDIO_LLM_BASE_URL")
)

AUDIO_TRANSCRIBE = (
    "Input is a base64 encoded audio. Transcribe the audio content. "
    "Return a json string with the following format: "
    '{"audio_text": "transcribed text from audio"}'
)


def encode_audio(audio_source: str, with_header: bool = True) -> str:
    """
    Encode audio to base64 format with robust file handling

    Args:
        audio_source: URL or local file path of the audio
        with_header: Whether to include MIME type header

    Returns:
        str: Base64 encoded audio string, with MIME type prefix if with_header is True

    Raises:
        ValueError: When audio source is invalid or audio format is not supported
        IOError: When audio file cannot be read
    """
    if not audio_source:
        raise ValueError("Audio source cannot be empty")

    try:
        # Get file with validation (only audio files allowed)
        file_path, mime_type, content = get_file_from_source(
            audio_source,
            allowed_mime_prefixes=["audio/"],
            max_size_mb=50.0,  # 50MB limit for audio files
            type="audio",  # Specify type as audio to handle audio files
        )

        # Encode to base64
        audio_base64 = base64.b64encode(content).decode()

        # Format with header if requested
        final_audio = (
            f"data:{mime_type};base64,{audio_base64}" if with_header else audio_base64
        )

        # Clean up temporary file if it was created for a URL
        if file_path != os.path.abspath(audio_source) and os.path.exists(file_path):
            os.unlink(file_path)

        return final_audio

    except Exception:
        logger.error(
            f"Error encoding audio from {audio_source}: {traceback.format_exc()}"
        )
        raise


@mcp.tool(description="Transcribe the given audio in a list of filepaths or urls.")
async def mcp_transcribe_audio(
    audio_urls: List[str] = Field(
        description="The input audio in given a list of filepaths or urls."
    ),
) -> str:
    """
    Transcribe the given audio in a list of filepaths or urls.

    Args:
        audio_urls: List of audio file paths or URLs

    Returns:
        str: JSON string containing transcriptions
    """
    transcriptions = []
    for audio_url in audio_urls:
        try:
            # Get file with validation (only audio files allowed)
            file_path, _, _ = get_file_from_source(
                audio_url,
                allowed_mime_prefixes=["audio/"],
                max_size_mb=50.0,  # 50MB limit for audio files
                type="audio",  # Specify type as audio to handle audio files
            )

            # Use the file for transcription
            with open(file_path, "rb") as audio_file:
                transcription = client.audio.transcriptions.create(
                    file=audio_file,
                    model=os.getenv("AUDIO_LLM_MODEL_NAME"),
                    response_format="text",
                )
                transcriptions.append(transcription)

            # Clean up temporary file if it was created for a URL
            if file_path != os.path.abspath(audio_url) and os.path.exists(file_path):
                os.unlink(file_path)

        except Exception as e:
            logger.error(f"Error transcribing {audio_url}: {traceback.format_exc()}")
            transcriptions.append(f"Error: {str(e)}")

    logger.info(f"---get_text_by_transcribe-transcription:{transcriptions}")
    return json.dumps(transcriptions, ensure_ascii=False)


def main():
    from dotenv import load_dotenv

    load_dotenv()

    print("Starting Audio MCP Server...", file=sys.stderr)
    mcp.run(transport="stdio")


# Make the module callable
def __call__():
    """
    Make the module callable for uvx.
    This function is called when the module is executed directly.
    """
    main()


# Add this for compatibility with uvx
import sys

sys.modules[__name__].__call__ = __call__

# Run the server when the script is executed directly
if __name__ == "__main__":
    main()



# coding: utf-8
# Copyright (c) 2025 inclusionAI.
import copy
import inspect
import os.path
from typing import Dict, Any, List, Union

from aworld.core.context.base import Context
from aworld.logs.util import logger
from aworld.utils import import_package


def usage_process(usage: Dict[str, Union[int, Dict[str, int]]] = {}, context: Context = None):
    if not context:
        context = Context.instance()

    stacks = inspect.stack()
    index = 0
    for idx, stack in enumerate(stacks):
        index = idx + 1
        file = os.path.basename(stack.filename)
        # supported use `llm.py` utility function only
        if 'call_llm_model' in stack.function and file == 'llm.py':
            break

    if index >= len(stacks):
        logger.warning("not category usage find to count")
    else:
        instance = stacks[index].frame.f_locals.get('self')
        name = getattr(instance, "_name", "unknown")
        usage[name] = copy.copy(usage)
    # total usage
    context.add_token(usage)


def num_tokens_from_messages(messages, model="gpt-4o"):
    """Return the number of tokens used by a list of messages."""
    import_package("tiktoken")
    import tiktoken

    try:
        encoding = tiktoken.encoding_for_model(model)
    except KeyError:
        logger.warning(f"{model} model not found. Using cl100k_base encoding.")
        encoding = tiktoken.get_encoding("cl100k_base")

    tokens_per_message = 3
    tokens_per_name = 1

    num_tokens = 0
    for message in messages:
        num_tokens += tokens_per_message
        for key, value in message.items():
            num_tokens += len(encoding.encode(value))
            if key == "name":
                num_tokens += tokens_per_name
    num_tokens += 3
    return num_tokens


def agent_desc_transform(agent_dict: Dict[str, Any],
                         agents: List[str] = None,
                         provider: str = 'openai',
                         strategy: str = 'min') -> List[Dict[str, Any]]:
    """Default implement transform framework standard protocol to openai protocol of agent description.

    Args:
        agent_dict: Dict of descriptions of agents that are registered in the agent factory.
        agents: Description of special agents to use.
        provider: Different descriptions formats need to be processed based on the provider.
        strategy: The value is `min` or `max`, when no special agents are provided, `min` indicates no content returned,
                 `max` means get all agents' descriptions.
    """
    agent_as_tools = []
    if not agents and strategy == 'min':
        return agent_as_tools

    if provider and 'openai' in provider:
        for agent_name, agent_info in agent_dict.items():
            if agents and agent_name not in agents:
                logger.debug(f"{agent_name} can not supported in {agents}, you can set `tools` params to support it.")
                continue

            for action in agent_info["abilities"]:
                # Build parameter properties
                properties = {}
                required = []
                for param_name, param_info in action["params"].items():
                    properties[param_name] = {
                        "description": param_info["desc"],
                        "type": param_info["type"] if param_info["type"] != "str" else "string"
                    }
                    if param_info.get("required", False):
                        required.append(param_name)

                openai_function_schema = {
                    "name": f'{agent_name}__{action["name"]}',
                    "description": action["desc"],
                    "parameters": {
                        "type": "object",
                        "properties": properties,
                        "required": required
                    }
                }

                agent_as_tools.append({
                    "type": "function",
                    "function": openai_function_schema
                })
    return agent_as_tools


def tool_desc_transform(tool_dict: Dict[str, Any],
                        tools: List[str] = None,
                        black_tool_actions: Dict[str, List[str]] = {},
                        provider: str = 'openai',
                        strategy: str = 'min') -> List[Dict[str, Any]]:
    """Default implement transform framework standard protocol to openai protocol of tool description.

    Args:
        tool_dict: Dict of descriptions of tools that are registered in the agent factory.
        tools: Description of special tools to use.
        provider: Different descriptions formats need to be processed based on the provider.
        strategy: The value is `min` or `max`, when no special tools are provided, `min` indicates no content returned,
                 `max` means get all tools' descriptions.
    """
    openai_tools = []
    if not tools and strategy == 'min':
        return openai_tools

    if black_tool_actions is None:
        black_tool_actions = {}

    if provider and 'openai' in provider:
        for tool_name, tool_info in tool_dict.items():
            if tools and tool_name not in tools:
                logger.debug(f"{tool_name} can not supported in {tools}, you can set `tools` params to support it.")
                continue

            black_actions = black_tool_actions.get(tool_name, [])
            for action in tool_info["actions"]:
                if action['name'] in black_actions:
                    continue
                # Build parameter properties
                properties = {}
                required = []
                for param_name, param_info in action["params"].items():
                    properties[param_name] = {
                        "description": param_info["desc"],
                        "type": param_info["type"] if param_info["type"] != "str" else "string"
                    }
                    if param_info.get("required", False):
                        required.append(param_name)

                openai_function_schema = {
                    "name": f'{tool_name}__{action["name"]}',
                    "description": action["desc"],
                    "parameters": {
                        "type": "object",
                        "properties": properties,
                        "required": required
                    }
                }

                openai_tools.append({
                    "type": "function",
                    "function": openai_function_schema
                })
    return openai_tools



import re
from typing import Tuple, List
from aworld.logs.util import logger
from aworld.trace.base import Propagator, Carrier, TraceContext


class W3CTraceContextPropagator(Propagator):
    """
    OtelPropagator is a Propagator that extracts and injects using w3c TraceContext's headers.
        carrier = {
            "traceparent": "00-0af7651916cd43dd8448eb211c80319c-00f067aa0ba902b7-01",
            "tracestate": "congo=t61rcWkgMzE",
            "baggage": "key1=value1,key2=value2"
        }
    """
    _STATE_KEY_FORMAT = (
        r"[a-z][_0-9a-z\-\*\/]{0,255}|"
        r"[a-z0-9][_0-9a-z\-\*\/]{0,240}@[a-z][_0-9a-z\-\*\/]{0,13}"
    )
    _STATE_VALUE_FORMAT = (
        r"[\x20-\x2b\x2d-\x3c\x3e-\x7e]{0,255}[\x21-\x2b\x2d-\x3c\x3e-\x7e]"
    )
    _state_delimiter_pattern = re.compile(r"[ \t]*,[ \t]*")
    _state_member_pattern = re.compile(
        f"({_STATE_KEY_FORMAT})(=)({_STATE_VALUE_FORMAT})[ \t]*")

    _TRACEPARENT_HEADER_NAME = "traceparent"
    _TRACESTATE_HEADER_NAME = "tracestate"
    _TRACEPARENT_HEADER_FORMAT = (
        "^[ \t]*([0-9a-f]{2})-([0-9a-f]{32})-([0-9a-f]{16})-([0-9a-f]{2})"
        + "(-.*)?[ \t]*$"
    )
    _TRACEPARENT_HEADER_FORMAT_RE = re.compile(_TRACEPARENT_HEADER_FORMAT)

    def extract(self, carrier: Carrier) -> TraceContext:
        """
        Extract trace context from carrier.
        Args:
            carrier: The carrier to extract trace context from.
        Returns:
            A dict of trace context.
        """
        header = carrier.get(self._TRACEPARENT_HEADER_NAME) or carrier.get(
            'HTTP_' + self._TRACEPARENT_HEADER_NAME.upper())

        if header is None:
            return None

        match = re.search(self._TRACEPARENT_HEADER_FORMAT_RE, header)
        if not match:
            return None

        version: str = match.group(1)
        trace_id: str = match.group(2)
        span_id: str = match.group(3)
        trace_flags: str = match.group(4)

        logger.info(
            f"extract trace_id: {trace_id}, span_id: {span_id}, trace_flags: {trace_flags}, version: {version}")

        if trace_id == "0" * 32 or span_id == "0" * 16:
            return None
        if version == "00":
            if match.group(5):  # type: ignore
                return None
        if version == "ff":
            return None

        state_header = carrier.get(self._TRACESTATE_HEADER_NAME) or carrier.get(
            'HTTP_' + self._TRACESTATE_HEADER_NAME.upper())
        return TraceContext(
            trace_id=trace_id,
            span_id=span_id,
            trace_flags=trace_flags,
            version=version,
            attributes=(self._extract_state_from_header(state_header))
        )

    def inject(self, trace_context: TraceContext, carrier: Carrier) -> None:
        """
        Inject trace context into carrier.
        Args:
            context: The trace context to inject.
            carrier: The carrier to inject trace context into.
        """
        attribute_copy = trace_context.attributes.copy()
        version: str = trace_context.version
        trace_flags: str = trace_context.trace_flags
        trace_id = trace_context.trace_id
        span_id = trace_context.span_id
        logger.info(
            f"inject trace_id: {trace_id}, span_id: {span_id}, trace_flags: {trace_flags}, version: {version}")
        if (not trace_id or trace_id == "0" * 32
                or not span_id or span_id == "0" * 16):
            return

        if isinstance(trace_id, int):
            trace_id = format(trace_id, "032x")
        if isinstance(span_id, int):
            span_id = format(span_id, "016x")
        traceparent_string = f"{version}-{trace_id}-{span_id}-{trace_flags}"
        carrier.set(self._TRACEPARENT_HEADER_NAME, traceparent_string)
        tracestate_string = ",".join(
            f"{key}={value}" for key, value in attribute_copy.items())
        if tracestate_string:
            carrier.set(self._TRACESTATE_HEADER_NAME, tracestate_string)

    def _extract_state_from_header(self, header: str) -> dict:
        """
        Extract state from header.
        Args:
            header: The header to extract state from.
        Returns:
            A dict of state.
        """
        if header is None:
            return {}
        state = {}
        members: List[str] = re.split(self._state_delimiter_pattern, header)
        for member in members:
            # empty members are valid, but no need to process further.
            if not member:
                continue
            match = self._state_member_pattern.fullmatch(member)
            if not match:
                logger.warning(
                    "Member doesn't match the w3c identifiers format {member}")
                return state
            groups: Tuple[str, ...] = match.groups()
            key, _eq, value = groups
            # duplicate keys are not legal in header
            if key in state:
                return state
            state[key] = value
        return state



# coding: utf-8
# Copyright (c) 2025 inclusionAI.
import time

from typing import AsyncGenerator

from aworld.core.common import TaskItem
from aworld.core.tool.base import Tool, AsyncTool

from aworld.core.event.base import Message, Constants
from aworld.core.task import TaskResponse
from aworld.logs.util import logger
from aworld.runners.handler.base import DefaultHandler
from aworld.runners.utils import TaskType


class DefaultTaskHandler(DefaultHandler):
    def __init__(self, runner: 'TaskEventRunner'):
        self.runner = runner

    async def handle(self, message: Message) -> AsyncGenerator[Message, None]:
        if message.category != Constants.TASK:
            return

        topic = message.topic
        task_item: TaskItem = message.payload
        if topic == TaskType.SUBSCRIBE_TOOL:
            new_tools = message.payload.data
            for name, tool in new_tools.items():
                if isinstance(tool, Tool) or isinstance(tool, AsyncTool):
                    await self.runner.event_mng.register(Constants.TOOL, name, tool.step)
                    logger.info(f"dynamic register {name} tool.")
                else:
                    logger.warning(f"Unknown tool instance: {tool}")
            return
        elif topic == TaskType.SUBSCRIBE_AGENT:
            return
        elif topic == TaskType.ERROR:
            if task_item.stop:
                await self.runner.stop()
                logger.warning(f"task {self.runner.task.id} stop, cause: {task_item.msg}")
                self.runner._task_response = TaskResponse(msg=task_item.msg,
                                                          answer='',
                                                          success=False,
                                                          id=self.runner.task.id,
                                                          time_cost=(time.time() - self.runner.start_time),
                                                          usage=self.runner.context.token_usage)
                return
            # restart
            logger.warning(f"The task {self.runner.task.id} will be restarted due to error: {task_item.msg}.")
            yield Message(
                category=Constants.TASK,
                payload='',
                sender=self.name(),
                session_id=self.runner.context.session_id,
                topic=TaskType.START
            )
        elif topic == TaskType.FINISHED:
            self.runner._task_response = TaskResponse(answer=str(message.payload),
                                                      success=True,
                                                      id=self.runner.task.id,
                                                      time_cost=(time.time() - self.runner.start_time),
                                                      usage=self.runner.context.token_usage)
            await self.runner.stop()

            # clean sandbox
            swarm = self.runner.task.swarm
            if swarm and hasattr(swarm, 'agents') and swarm.agents:
                for agent_name, agent in swarm.agents.items():
                    try:
                        if hasattr(agent, 'sandbox') and agent.sandbox:
                            await agent.sandbox.cleanup()
                    except Exception as e:
                        logger.warning(f"Failed to cleanup sandbox for agent {agent_name}: {e}")

            logger.info(f"{self.runner.task.id} finished.")
            return
        elif topic == TaskType.START:
            logger.info(f"task start event: {message}, will send init message.")
            if message.payload:
                yield message
            else:
                yield self.runner.init_message
        elif topic == TaskType.OUTPUT:
            yield message



from typing import Dict, List, Any, Optional

from aworld.sandbox.base import Sandbox
from aworld.sandbox.common import BaseSandbox
from aworld.sandbox.models import SandboxEnvType
from aworld.sandbox.implementations import LocalSandbox, KubernetesSandbox, SuperSandbox


# For backward compatibility, use LocalSandbox as the default Sandbox implementation
DefaultSandbox = LocalSandbox


# Override Sandbox class constructor to create the appropriate sandbox based on env_type
def create_sandbox(
    env_type: Optional[int] = None,
    sandbox_id: Optional[str] = None,
    metadata: Optional[Dict[str, str]] = None,
    timeout: Optional[int] = None,
    mcp_servers: Optional[List[str]] = None,
    mcp_config: Optional[Any] = None,
    **kwargs
) -> Sandbox:
    """
    Factory function to create a sandbox instance based on the environment type.
    
    Args:
        env_type: The environment type. Defaults to LOCAL if None.
        sandbox_id: Unique identifier for the sandbox. If None, one will be generated.
        metadata: Additional metadata for the sandbox.
        timeout: Timeout for sandbox operations.
        mcp_servers: List of MCP servers to use.
        mcp_config: Configuration for MCP servers.
        **kwargs: Additional parameters for specific sandbox types.
        
    Returns:
        Sandbox: An instance of a sandbox implementation.
        
    Raises:
        ValueError: If an invalid environment type is provided.
    """
    env_type = env_type or SandboxEnvType.LOCAL
    
    if env_type == SandboxEnvType.LOCAL:
        return LocalSandbox(
            sandbox_id=sandbox_id,
            metadata=metadata,
            timeout=timeout,
            mcp_servers=mcp_servers,
            mcp_config=mcp_config,
            **kwargs
        )
    elif env_type == SandboxEnvType.K8S:
        return KubernetesSandbox(
            sandbox_id=sandbox_id,
            metadata=metadata,
            timeout=timeout,
            mcp_servers=mcp_servers,
            mcp_config=mcp_config,
            **kwargs
        )
    elif env_type == SandboxEnvType.SUPERCOMPUTER:
        return SuperSandbox(
            sandbox_id=sandbox_id,
            metadata=metadata,
            timeout=timeout,
            mcp_servers=mcp_servers,
            mcp_config=mcp_config,
            **kwargs
        )
    else:
        raise ValueError(f"Invalid environment type: {env_type}")


# Monkey patch the Sandbox class to make direct instantiation work
old_init = Sandbox.__init__

def _sandbox_init(self, *args, **kwargs):
    if type(self) is Sandbox:
        # This should never be called directly, as __new__ will return a different type
        pass
    else:
        # Pass through to the original __init__ for actual implementations
        old_init(self, *args, **kwargs)

# Store the original __new__ method
original_new = object.__new__

# Create a new __new__ method that intercepts Sandbox instantiation
def _sandbox_new(cls, *args, **kwargs):
    if cls is Sandbox:
        # If trying to instantiate Sandbox directly, use our factory instead
        return create_sandbox(**kwargs)
    else:
        # For subclasses, use the original __new__
        return original_new(cls)

# Apply the monkey patches
Sandbox.__init__ = _sandbox_init
Sandbox.__new__ = _sandbox_new


# Expose key classes and functions
__all__ = [
    'Sandbox',
    'BaseSandbox',
    'LocalSandbox',
    'KubernetesSandbox',
    'SuperSandbox',
    'DefaultSandbox',
    'SandboxEnvType',
    'create_sandbox'
]



# coding: utf-8
# Copyright (c) 2025 inclusionAI.
import threading
from .routes import setup_routes
from aworld.logs.util import logger

GLOBAL_TRACE_SERVER = None


class TraceServer:
    def __init__(self, storage, port: int = 7079):
        self._storage = storage
        self._port = port
        self._thread = None
        self.app = None

    def start(self):
        self._thread = threading.Thread(target=self._start_app, daemon=True)
        self._thread.start()

    def join(self):
        if self._thread:
            self._thread.join()

    def _start_app(self):
        app = setup_routes(self._storage)
        self.app = app
        app.run(port=self._port)


def start_trace_server(storage, port: int = 8000):
    global GLOBAL_TRACE_SERVER
    if GLOBAL_TRACE_SERVER is not None:
        setup_routes(storage)
    else:
        GLOBAL_TRACE_SERVER = TraceServer(storage, port)
        GLOBAL_TRACE_SERVER.start()


def get_trace_server():
    if GLOBAL_TRACE_SERVER is None:
        logger.warning("No trace server has been set.")
    return GLOBAL_TRACE_SERVER



import os
from typing import Any, Dict, List, Generator, AsyncGenerator

from aworld import import_package
from aworld.logs.util import logger
from aworld.core.llm_provider_base import LLMProviderBase
from aworld.models.model_response import ModelResponse, LLMResponseError


class AnthropicProvider(LLMProviderBase):
    """Anthropic provider implementation.
    """

    def __init__(self,
                 api_key: str = None,
                 base_url: str = None,
                 model_name: str = None,
                 sync_enabled: bool = None,
                 async_enabled: bool = None,
                 **kwargs):
        super().__init__(api_key, base_url, model_name, sync_enabled, async_enabled, **kwargs)
        import_package("anthropic")

    def _init_provider(self):
        """Initialize Anthropic provider.

        Returns:
            Anthropic provider instance.
        """
        from anthropic import Anthropic

        # Get API key
        api_key = self.api_key
        if not api_key:
            env_var = "ANTHROPIC_API_KEY"
            api_key = os.getenv(env_var, "")
            if not api_key:
                raise ValueError(
                    f"Anthropic API key not found, please set {env_var} environment variable or provide it in the parameters")

        return Anthropic(
            api_key=api_key,
            base_url=self.base_url
        )

    def _init_async_provider(self):
        """Initialize async Anthropic provider.

        Returns:
            Async Anthropic provider instance.
        """
        from anthropic import Anthropic, AsyncAnthropic

        # Get API key
        api_key = self.api_key
        if not api_key:
            env_var = "ANTHROPIC_API_KEY"
            api_key = os.getenv(env_var, "")
            if not api_key:
                raise ValueError(
                    f"Anthropic API key not found, please set {env_var} environment variable or provide it in the parameters")

        return AsyncAnthropic(
            api_key=api_key,
            base_url=self.base_url
        )

    @classmethod
    def supported_models(cls) -> list[str]:
        return [r"claude-3-.*"]

    def preprocess_messages(self, messages: List[Dict[str, str]]) -> Dict[str, Any]:
        """Preprocess messages, convert OpenAI format to Anthropic format.

        Args:
            messages: OpenAI format message list.

        Returns:
            Converted message dictionary, containing messages and system fields.
        """
        anthropic_messages = []
        system_content = None

        for msg in messages:
            role = msg.get("role", "")
            content = msg.get("content", "")

            if role == "system":
                system_content = content
            elif role == "user":
                anthropic_messages.append({"role": "user", "content": content})
            elif role == "assistant":
                anthropic_messages.append({"role": "assistant", "content": content})

        return {
            "messages": anthropic_messages,
            "system": system_content
        }

    def postprocess_response(self, response: Any) -> ModelResponse:
        """Process Anthropic response to unified ModelResponse.

        Args:
            response: Anthropic response object.

        Returns:
            ModelResponse object.
            
        Raises:
            LLMResponseError: When LLM response error occurs.
        """
        # Check if response is empty or contains error
        if not response or (isinstance(response, dict) and response.get('error')):
            error_msg = response.get('error', 'Unknown error') if isinstance(response, dict) else 'Empty response'
            raise LLMResponseError(error_msg, self.model_name or "claude", response)

        return ModelResponse.from_anthropic_response(response)

    def postprocess_stream_response(self, chunk: Any) -> ModelResponse:
        """Process Anthropic streaming response chunk.

        Args:
            chunk: Anthropic response chunk.

        Returns:
            ModelResponse object.
            
        Raises:
            LLMResponseError: When LLM response error occurs.
        """
        # Check if chunk is empty or contains error
        if not chunk or (isinstance(chunk, dict) and chunk.get('error')):
            error_msg = chunk.get('error', 'Unknown error') if isinstance(chunk, dict) else 'Empty response'
            raise LLMResponseError(error_msg, self.model_name or "claude", chunk)

        return ModelResponse.from_anthropic_stream_chunk(chunk)

    def completion(self,
                   messages: List[Dict[str, str]],
                   temperature: float = 0.0,
                   max_tokens: int = None,
                   stop: List[str] = None,
                   **kwargs) -> ModelResponse:
        """Synchronously call Anthropic to generate response.
        
        Args:
            messages: Message list.
            temperature: Temperature parameter.
            max_tokens: Maximum number of tokens to generate.
            stop: List of stop sequences.
            **kwargs: Other parameters.

        Returns:
            ModelResponse object.
        """
        if not self.provider:
            raise RuntimeError(
                "Sync provider not initialized. Make sure 'sync_enabled' parameter is set to True in initialization.")

        try:
            processed_data = self.preprocess_messages(messages)
            processed_messages = processed_data["messages"]
            system_content = processed_data["system"]
            anthropic_params = self.get_anthropic_params(processed_messages, system_content, temperature, max_tokens,
                                                         stop, **kwargs)
            response = self.provider.visited_messages.create(**anthropic_params)

            return self.postprocess_response(response)
        except Exception as e:
            logger.warn(f"Error in Anthropic completion: {e}")
            raise LLMResponseError(str(e), kwargs.get("model_name", self.model_name or "claude"))

    def stream_completion(self,
                          messages: List[Dict[str, str]],
                          temperature: float = 0.0,
                          max_tokens: int = None,
                          stop: List[str] = None,
                          **kwargs) -> Generator[ModelResponse, None, None]:
        """Synchronously call Anthropic to generate streaming response.

        Args:
            messages: Message list.
            temperature: Temperature parameter.
            max_tokens: Maximum number of tokens to generate.
            stop: List of stop sequences.
            **kwargs: Other parameters.

        Returns:
            Generator yielding ModelResponse chunks.
        """
        if not self.provider:
            raise RuntimeError(
                "Sync provider not initialized. Make sure 'sync_enabled' parameter is set to True in initialization.")

        try:
            processed_data = self.preprocess_messages(messages)
            processed_messages = processed_data["messages"]
            system_content = processed_data["system"]
            anthropic_params = self.get_anthropic_params(processed_messages, system_content, temperature, max_tokens,
                                                         stop, **kwargs)
            anthropic_params["stream"] = True
            response_stream = self.provider.visited_messages.create(**anthropic_params)

            for chunk in response_stream:
                if not chunk:
                    continue

                yield self.postprocess_stream_response(chunk)

        except Exception as e:
            logger.warn(f"Error in Anthropic stream_completion: {e}")
            raise LLMResponseError(str(e), kwargs.get("model_name", self.model_name or "claude"))

    async def astream_completion(self,
                                 messages: List[Dict[str, str]],
                                 temperature: float = 0.0,
                                 max_tokens: int = None,
                                 stop: List[str] = None,
                                 **kwargs) -> AsyncGenerator[ModelResponse, None]:
        """Asynchronously call Anthropic to generate streaming response.
        
        Args:
            messages: Message list.
            temperature: Temperature parameter.
            max_tokens: Maximum number of tokens to generate.
            stop: List of stop sequences.
            **kwargs: Other parameters.

        Returns:
            AsyncGenerator yielding ModelResponse chunks.
        """
        if not self.async_provider:
            raise RuntimeError(
                "Async provider not initialized. Make sure 'async_enabled' parameter is set to True in initialization.")

        try:
            processed_data = self.preprocess_messages(messages)
            processed_messages = processed_data["messages"]
            system_content = processed_data["system"]
            anthropic_params = self.get_anthropic_params(processed_messages, system_content, temperature, max_tokens,
                                                         stop, **kwargs)
            anthropic_params["stream"] = True
            response_stream = await self.async_provider.visited_messages.create(**anthropic_params)

            async for chunk in response_stream:
                if not chunk:
                    continue

                yield self.postprocess_stream_response(chunk)

        except Exception as e:
            logger.warn(f"Error in Anthropic astream_completion: {e}")
            raise LLMResponseError(str(e), kwargs.get("model_name", self.model_name or "claude"))

    async def acompletion(self,
                          messages: List[Dict[str, str]],
                          temperature: float = 0.0,
                          max_tokens: int = None,
                          stop: List[str] = None,
                          **kwargs) -> ModelResponse:
        """Asynchronously call Anthropic to generate response.
        
        Args:
            messages: Message list.
            temperature: Temperature parameter.
            max_tokens: Maximum number of tokens to generate.
            stop: List of stop sequences.
            **kwargs: Other parameters.

        Returns:
            ModelResponse object.
        """
        if not self.async_provider:
            raise RuntimeError(
                "Async provider not initialized. Make sure 'async_enabled' parameter is set to True in initialization.")

        try:
            processed_data = self.preprocess_messages(messages)
            processed_messages = processed_data["messages"]
            system_content = processed_data["system"]
            anthropic_params = self.get_anthropic_params(processed_messages, system_content, temperature, max_tokens,
                                                         stop, **kwargs)
            response = await self.async_provider.visited_messages.create(**anthropic_params)

            return self.postprocess_response(response)
        except Exception as e:
            logger.warn(f"Error in Anthropic acompletion: {e}")
            raise LLMResponseError(str(e), kwargs.get("model_name", self.model_name or "claude"))

    def get_anthropic_params(self,
                             messages: List[Dict[str, str]],
                             system: str = None,
                             temperature: float = 0.0,
                             max_tokens: int = None,
                             stop: List[str] = None,
                             **kwargs) -> Dict[str, Any]:
        if "tools" in kwargs:
            openai_tools = kwargs["tools"]
            claude_tools = []

            for tool in openai_tools:
                if tool["type"] == "function":
                    claude_tool = {
                        "name": tool["name"],
                        "description": tool["description"],
                        "input_schema": {
                            "type": "object",
                            "properties": tool["parameters"]["properties"],
                            "required": tool["parameters"].get("required", [])
                        }
                    }
                    claude_tools.append(claude_tool)

            kwargs["tools"] = claude_tools

        anthropic_params = {
            "model": kwargs.get("model_name", self.model_name or ""),
            "messages": messages,
            "system": system,
            "temperature": temperature,
            "max_tokens": max_tokens or 4096,
            "stop_sequences": stop,
        }

        if "tools" in kwargs and kwargs["tools"]:
            anthropic_params["tools"] = kwargs["tools"]
            anthropic_params["tool_choice"] = kwargs.get("tool_choice", "auto")

        for param in ["top_p", "top_k", "metadata", "stream"]:
            if param in kwargs:
                anthropic_params[param] = kwargs[param]

        return anthropic_params



import ast
import re
import sys
import warnings
from importlib.abc import Loader, MetaPathFinder
from importlib.machinery import ModuleSpec
from importlib.util import spec_from_loader
from types import ModuleType
from typing import TYPE_CHECKING, Sequence, Union, Callable, Iterator, TypeVar, Any, cast

from aworld.trace.base import log_trace_error
from .rewrite_ast import compile_source

if TYPE_CHECKING:
    from .context_manager import TraceManager


class AutoTraceModule:
    """A class that represents a module being imported that should maybe be traced automatically."""

    def __init__(self, module_name: str) -> None:
        self._module_name = module_name
        """Fully qualified absolute name of the module being imported."""

    def need_auto_trace(self, prefix: Union[str, Sequence[str]]) -> bool:
        """
        Check if the module name starts with the given prefix.
        """
        if isinstance(prefix, str):
            prefix = (prefix,)
        pattern = '|'.join([get_module_pattern(p) for p in prefix])
        return bool(re.match(pattern, self._module_name))


class TraceImportFinder(MetaPathFinder):
    """A class that implements the `find_spec` method of the `MetaPathFinder` protocol."""

    def __init__(self, trace_manager: "TraceManager", module_funcs: Callable[[AutoTraceModule], bool],
                 min_duration_ns: int) -> None:
        self._trace_manager = trace_manager
        self._modules_filter = module_funcs
        self._min_duration_ns = min_duration_ns

    def _find_plain_specs(
            self, fullname: str, path: Sequence[str] = None, target: ModuleType = None
    ) -> Iterator[ModuleSpec]:
        """Yield module specs returned by other finders on `sys.meta_path`."""
        for finder in sys.meta_path:
            # Skip this finder or any like it to avoid infinite recursion.
            if isinstance(finder, TraceImportFinder):
                continue

            try:
                plain_spec = finder.find_spec(fullname, path, target)
            except Exception:  # pragma: no cover
                continue

            if plain_spec:
                yield plain_spec

    def find_spec(self, fullname: str, path: Sequence[str], target=None) -> None:
        """Find the spec for the given module name."""

        for plain_spec in self._find_plain_specs(fullname, path, target):
            # Get module specs returned by other finders on `sys.meta_path`
            get_source = getattr(plain_spec.loader, 'get_source', None)
            if not callable(get_source):
                continue
            try:
                source = cast(str, get_source(fullname))
            except Exception:
                continue

            if not source:
                continue

            filename = plain_spec.origin
            if not filename:
                try:
                    filename = cast('str | None', plain_spec.loader.get_filename(fullname))
                except Exception:
                    pass
            filename = filename or f'<{fullname}>'

            if not self._modules_filter(AutoTraceModule(fullname)):
                return None

            try:
                tree = ast.parse(source)
            except Exception:
                # Invalid source code. Try another one.
                continue

            try:
                execute = compile_source(tree, filename, fullname, self._trace_manager, self._min_duration_ns)
            except Exception:  # pragma: no cover
                log_trace_error()
                return None

            loader = AutoTraceLoader(plain_spec, execute)
            return spec_from_loader(fullname, loader)


class AutoTraceLoader(Loader):
    """
    A class that implements the `exec_module` method of the `Loader` protocol.
    """

    def __init__(self, plain_spec: ModuleSpec, execute: Callable[[dict[str, Any]], None]) -> None:
        self._plain_spec = plain_spec
        self._execute = execute

    def exec_module(self, module: ModuleType):
        """Execute a modified AST of the module's source code in the module's namespace.
        """
        self._execute(module.__dict__)

    def create_module(self, spec: ModuleSpec):
        return None

    def get_code(self, _name: str):
        """`python -m` uses the `runpy` module which calls this method instead of going through the normal protocol.
        So return some code which can be executed with the module namespace.
        Here `__loader__` will be this object, i.e. `self`.
        source = '__loader__.execute(globals())'
        return compile(source, '<string>', 'exec', dont_inherit=True)
        """

    def __getattr__(self, item: str):
        """Forward some methods to the plain spec's loader (likely a `SourceFileLoader`) if they exist."""
        if item in {'get_filename', 'is_package'}:
            return getattr(self.plain_spec.loader, item)
        raise AttributeError(item)


def convert_to_modules_func(modules: Sequence[str]) -> Callable[[AutoTraceModule], bool]:
    """Convert a sequence of module names to a function that checks if a module name starts with any of the given module names.
    """
    return lambda module: module.need_auto_trace(modules)


def get_module_pattern(module: str):
    """
    Get the regex pattern for the given module name.
    """

    if not re.match(r'[\w.]+$', module, re.UNICODE):
        return module
    module = re.escape(module)
    return rf'{module}($|\.)'


def install_auto_tracing(trace_manager: "TraceManager",
                         modules: Union[Sequence[str],
                         Callable[[AutoTraceModule], bool]],
                         min_duration_seconds: float
                         ) -> None:
    """
    Automatically trace the execution of a function.
    """
    if isinstance(modules, Sequence):
        module_funcs = convert_to_modules_func(modules)
    else:
        module_funcs = modules

    if not callable(module_funcs):
        raise TypeError('modules must be a list of strings or a callable')

    for module in list(sys.modules.values()):
        try:
            auto_trace_module = AutoTraceModule(module.__name__)
        except Exception:
            continue

        if module_funcs(auto_trace_module):
            warnings.warn(f'The module {module.__name__!r} matches modules to trace, but it has already been imported. '
                          f'Call `auto_tracing` earlier',
                          stacklevel=2,
                          )

    min_duration_ns = int(min_duration_seconds * 1_000_000_000)
    trace_manager = trace_manager.new_manager('auto_tracing')
    finder = TraceImportFinder(trace_manager, module_funcs, min_duration_ns)
    sys.meta_path.insert(0, finder)


T = TypeVar('T')


def not_auto_trace(x: T) -> T:
    """Decorator to prevent a function/class from being traced by `auto_tracing`"""
    return x



import flask
import weakref
from typing import Any, Callable, Collection
from time import time_ns
from timeit import default_timer
from importlib_metadata import version
from packaging import version as package_version
from aworld.trace.instrumentation import Instrumentor
from aworld.trace.base import Span, TraceProvider, TraceContext, Tracer, SpanType, get_tracer_provider
from aworld.metrics.metric import MetricType
from aworld.metrics.template import MetricTemplate
from aworld.logs.util import logger
from aworld.trace.instrumentation.http_util import (
    collect_request_attributes,
    url_disabled,
    get_excluded_urls,
    parse_excluded_urls,
    HTTP_ROUTE
)
from aworld.trace.propagator import get_global_trace_propagator
from aworld.metrics.context_manager import MetricContext
from aworld.trace.propagator.carrier import ListTupleCarrier, DictCarrier

_ENVIRON_STARTTIME_KEY = "aworld-flask.starttime_key"
_ENVIRON_SPAN_KEY = "aworld-flask.span_key"
_ENVIRON_REQCTX_REF_KEY = "aworld-flask.reqctx_ref_key"

flask_version = version("flask")
if package_version.parse(flask_version) >= package_version.parse("2.2.0"):

    def _request_ctx_ref() -> weakref.ReferenceType:
        return weakref.ref(flask.globals.request_ctx._get_current_object())

else:

    def _request_ctx_ref() -> weakref.ReferenceType:
        return weakref.ref(flask._request_ctx_stack.top)


def _rewrapped_app(
    wsgi_app,
    active_requests_counter,
    duration_histogram,
    response_hook=None,
    excluded_urls=None,
):
    def _wrapped_app(wrapped_app_environ, start_response):
        # We want to measure the time for route matching, etc.
        # In theory, we could start the span here and use
        # update_name later but that API is "highly discouraged" so
        # we better avoid it.
        wrapped_app_environ[_ENVIRON_STARTTIME_KEY] = time_ns()
        start = default_timer()
        attributes = collect_request_attributes(wrapped_app_environ)

        if MetricContext.metric_initialized():
            MetricContext.inc(active_requests_counter, 1, attributes)

        request_route = None

        def _start_response(status, response_headers, *args, **kwargs):
            if flask.request and (
                excluded_urls is None
                or not url_disabled(flask.request.url, excluded_urls)
            ):
                nonlocal request_route
                request_route = flask.request.url_rule

                span: Span = flask.request.environ.get(_ENVIRON_SPAN_KEY)

                propagator = get_global_trace_propagator()
                if propagator and span:
                    trace_context = TraceContext(
                        trace_id=span.get_trace_id(),
                        span_id=span.get_span_id()
                    )
                    propagator.inject(
                        trace_context, ListTupleCarrier(response_headers))

                if span and span.is_recording():
                    status_code_str, _ = status.split(" ", 1)
                    try:
                        status_code = int(status_code_str)
                    except ValueError:
                        status_code = -1

                    span.set_attribute(
                        "http.response.status_code", status_code)
                    span.set_attributes(attributes)

                if response_hook is not None:
                    response_hook(span, status, response_headers)
            return start_response(status, response_headers, *args, **kwargs)

        result = wsgi_app(wrapped_app_environ, _start_response)
        duration_s = default_timer() - start

        if MetricContext.metric_initialized():
            MetricContext.histogram_record(
                duration_histogram,
                duration_s,
                attributes
            )
            MetricContext.dec(active_requests_counter, 1, attributes)
        return result

    return _wrapped_app


def _wrapped_before_request(
    request_hook=None,
    tracer: Tracer = None,
    excluded_urls=None
):
    def _before_request():
        if excluded_urls and url_disabled(flask.request.url, excluded_urls):
            return
        flask_request_environ = flask.request.environ
        logger.info(
            f"_wrapped_before_request flask_request_environ={flask_request_environ}")

        attributes = collect_request_attributes(flask_request_environ)

        if flask.request.url_rule:
            # For 404 that result from no route found, etc, we
            # don't have a url_rule.
            attributes[HTTP_ROUTE] = flask.request.url_rule.rule
            span_name = f"HTTP {flask.request.url_rule.rule}"
        else:
            span_name = f"HTTP {flask.request.url}"

        propagator = get_global_trace_propagator()
        trace_context = None
        if propagator:
            trace_context = propagator.extract(
                DictCarrier(flask_request_environ))

        logger.info(f"_wrapped_before_request trace_context={trace_context}")

        span = tracer.start_span(
            span_name,
            SpanType.SERVER,
            attributes=attributes,
            start_time=flask_request_environ.get(_ENVIRON_STARTTIME_KEY),
            trace_context=trace_context
        )

        if request_hook:
            request_hook(span, flask_request_environ)

        flask_request_environ[_ENVIRON_SPAN_KEY] = span
        flask_request_environ[_ENVIRON_REQCTX_REF_KEY] = _request_ctx_ref()

    return _before_request


def _wrapped_teardown_request(
    excluded_urls=None,
):
    def _teardown_request(exc):
        if excluded_urls and url_disabled(flask.request.url, excluded_urls):
            return

        span: Span = flask.request.environ.get(_ENVIRON_SPAN_KEY)

        original_reqctx_ref = flask.request.environ.get(
            _ENVIRON_REQCTX_REF_KEY
        )
        current_reqctx_ref = _request_ctx_ref()
        if not span or original_reqctx_ref != current_reqctx_ref:
            return
        if exc is None:
            span.end()
        else:
            span.record_exception(exc)
            span.end()

    return _teardown_request


class _InstrumentedFlask(flask.Flask):
    _excluded_urls = None
    _tracer_provider: TraceProvider = None
    _request_hook = None
    _response_hook = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        tracer = self._tracer_provider.get_tracer(
            "aworld.trace.instrumentation.flask")

        duration_histogram = MetricTemplate(
            type=MetricType.HISTOGRAM,
            name="flask_request_duration_histogram",
            description="Duration of flask HTTP server requests."
        )

        active_requests_counter = MetricTemplate(
            type=MetricType.UPDOWNCOUNTER,
            name="flask_active_request_counter",
            unit="1",
            description="Number of active HTTP server requests.",
        )

        self.wsgi_app = _rewrapped_app(
            self.wsgi_app,
            active_requests_counter,
            duration_histogram,
            _InstrumentedFlask._response_hook,
            excluded_urls=_InstrumentedFlask._excluded_urls
        )

        _before_request = _wrapped_before_request(
            _InstrumentedFlask._request_hook,
            tracer,
            excluded_urls=_InstrumentedFlask._excluded_urls
        )
        self._before_request = _before_request
        self.before_request(_before_request)

        _teardown_request = _wrapped_teardown_request(
            excluded_urls=_InstrumentedFlask._excluded_urls,
        )
        self.teardown_request(_teardown_request)


class FlaskInstrumentor(Instrumentor):

    def instrumentation_dependencies(self) -> Collection[str]:
        return ("flask >= 1.0",)

    def _instrument(self, **kwargs: Any):
        logger.info("Flask _instrument entered.")
        self._original_flask = flask.Flask
        request_hook = kwargs.get("request_hook")
        response_hook = kwargs.get("response_hook")
        if callable(request_hook):
            _InstrumentedFlask._request_hook = request_hook
        if callable(response_hook):
            _InstrumentedFlask._response_hook = response_hook
        tracer_provider = kwargs.get("tracer_provider")
        _InstrumentedFlask._tracer_provider = tracer_provider
        excluded_urls = kwargs.get("excluded_urls")
        _InstrumentedFlask._excluded_urls = (
            get_excluded_urls("FLASK")
            if excluded_urls is None
            else parse_excluded_urls(excluded_urls)
        )
        flask.Flask = _InstrumentedFlask
        logger.info("Flask _instrument exited.")

    def _uninstrument(self, **kwargs):
        flask.Flask = self._original_flask


def instrument_flask(excluded_urls: str = None,
                     request_hook: Callable = None,
                     response_hook: Callable = None,
                     tracer_provider: TraceProvider = None,
                     **kwargs: Any,
                     ):
    """
    Instrument the Flask application.
    Args:
        excluded_urls (str): A comma separated list of URLs to be excluded from instrumentation.
        request_hook (Callable): A function to be called before a request is processed.
        response_hook (Callable): A function to be called after a request is processed.
        tracer_provider (TraceProvider): The trace provider to use.
    """
    all_kwargs = {
        "excluded_urls": excluded_urls,
        "request_hook": request_hook,
        "response_hook": response_hook,
        "tracer_provider": tracer_provider or get_tracer_provider(),
        **kwargs
    }
    FlaskInstrumentor().instrument(**all_kwargs)
    logger.info("Flask instrumented.")



import asyncio
import json
import os
import tempfile
from typing import List, Optional, Tuple
from urllib.parse import urlparse

import requests
from mcp.server import FastMCP

from aworld.logs.util import logger
from aworld.mcp_client.utils import mcp_tool_desc_transform


def get_mime_type(file_path: str, default_mime: Optional[str] = None) -> str:
    """
    Detect MIME type of a file using python-magic if available,
    otherwise fallback to extension-based detection.

    Args:
        file_path: Path to the file
        default_mime: Default MIME type to return if detection fails

    Returns:
        str: Detected MIME type
    """
    # Try using python-magic for accurate MIME type detection
    try:
        # mime = magic.Magic(mime=True)
        # return mime.from_file(file_path)
        return "audio/mpeg"
    except (AttributeError, IOError):
        # Fallback to extension-based detection
        extension_mime_map = {
            # Audio formats
            ".mp3": "audio/mpeg",
            ".wav": "audio/wav",
            ".ogg": "audio/ogg",
            ".m4a": "audio/mp4",
            ".flac": "audio/flac",
            # Image formats
            ".jpg": "image/jpeg",
            ".jpeg": "image/jpeg",
            ".png": "image/png",
            ".gif": "image/gif",
            ".webp": "image/webp",
            ".bmp": "image/bmp",
            ".tiff": "image/tiff",
            # Video formats
            ".mp4": "video/mp4",
            ".avi": "video/x-msvideo",
            ".mov": "video/quicktime",
            ".mkv": "video/x-matroska",
            ".webm": "video/webm",
        }

        ext = os.path.splitext(file_path)[1].lower()
        return extension_mime_map.get(ext, default_mime or "application/octet-stream")


def is_url(path_or_url: str) -> bool:
    """
    Check if the given string is a URL.

    Args:
        path_or_url: String to check

    Returns:
        bool: True if the string is a URL, False otherwise
    """
    parsed = urlparse(path_or_url)
    return bool(parsed.scheme and parsed.netloc)


def get_file_from_source(
    source: str,
    allowed_mime_prefixes: List[str] = None,
    max_size_mb: float = 100.0,
    timeout: int = 60,
    type: str = "image",
) -> Tuple[str, str, bytes]:
    """
    Unified function to get file content from a URL or local path with validation.

    Args:
        source: URL or local file path
        allowed_mime_prefixes: List of allowed MIME type prefixes (e.g., ['audio/', 'video/'])
        max_size_mb: Maximum allowed file size in MB
        timeout: Timeout for URL requests in seconds

    Returns:
        Tuple[str, str, bytes]: (file_path, mime_type, file_content)
        - For URLs, file_path will be a temporary file path
        - For local files, file_path will be the original path

    Raises:
        ValueError: When file doesn't exist, exceeds size limit, or has invalid MIME type
        IOError: When file cannot be read
        requests.RequestException: When URL request fails
    """
    max_size_bytes = max_size_mb * 1024 * 1024
    temp_file = None

    try:
        if is_url(source):
            # Handle URL
            logger.info(f"Downloading file from URL: {source}")
            response = requests.get(source, stream=True, timeout=timeout)
            response.raise_for_status()

            # Check Content-Length if available
            content_length = response.headers.get("Content-Length")
            if content_length and int(content_length) > max_size_bytes:
                raise ValueError(f"File size exceeds limit of {max_size_mb}MB")

            # Create a temporary file
            temp_file = tempfile.NamedTemporaryFile(delete=False)
            file_path = temp_file.name

            # Download content in chunks to avoid memory issues
            content = bytearray()
            downloaded_size = 0
            for chunk in response.iter_content(chunk_size=8192):
                downloaded_size += len(chunk)
                if downloaded_size > max_size_bytes:
                    raise ValueError(f"File size exceeds limit of {max_size_mb}MB")
                temp_file.write(chunk)
                content.extend(chunk)

            temp_file.close()

            # Get MIME type
            if type == "audio":
                mime_type = "audio/mpeg"
            elif type == "image":
                mime_type = "image/jpeg"
            elif type == "video":
                mime_type = "video/mp4"

            # mime_type = get_mime_type(file_path)

            # For URLs where magic fails, try to use Content-Type header
            if mime_type == "application/octet-stream":
                content_type = response.headers.get("Content-Type", "").split(";")[0]
                if content_type:
                    mime_type = content_type
        else:
            # Handle local file
            file_path = os.path.abspath(source)

            # Check if file exists
            if not os.path.exists(file_path):
                raise ValueError(f"File not found: {file_path}")

            # Check file size
            file_size = os.path.getsize(file_path)
            if file_size > max_size_bytes:
                raise ValueError(f"File size exceeds limit of {max_size_mb}MB")

            # Get MIME type
            if type == "audio":
                mime_type = "audio/mpeg"
            elif type == "image":
                mime_type = "image/jpeg"
            elif type == "video":
                mime_type = "video/mp4"
            # mime_type = get_mime_type(file_path)

            # Read file content
            with open(file_path, "rb") as f:
                content = f.read()

        # Validate MIME type if allowed_mime_prefixes is provided
        if allowed_mime_prefixes:
            if not any(
                mime_type.startswith(prefix) for prefix in allowed_mime_prefixes
            ):
                allowed_types = ", ".join(allowed_mime_prefixes)
                raise ValueError(
                    f"Invalid file type: {mime_type}. Allowed types: {allowed_types}"
                )

        return file_path, mime_type, content

    except Exception as e:
        # Clean up temporary file if an error occurs
        if temp_file and os.path.exists(temp_file.name):
            os.unlink(temp_file.name)
        raise e


if __name__ == "__main__":
    mcp_tools = asyncio.run(mcp_tool_desc_transform(["search"]))
    logger.success(f"{json.dumps(mcp_tools, indent=4, ensure_ascii=False)}")



"""
Document MCP Server

This module provides MCP server functionality for document processing and analysis.
It handles various document formats including:
- Text files
- PDF documents
- Word documents (DOCX)
- Excel spreadsheets
- PowerPoint presentations
- JSON and XML files
- Source code files

Each document type has specialized processing functions that extract content,
structure, and metadata. The server focuses on local file processing with
appropriate validation and error handling.

Main functions:
- mcpreadtext: Reads plain text files
- mcpreadpdf: Reads PDF files with optional image extraction
- mcpreaddocx: Reads Word documents
- mcpreadexcel: Reads Excel spreadsheets
- mcpreadpptx: Reads PowerPoint presentations
- mcpreadjson: Reads and parses JSON/JSONL files
- mcpreadxml: Reads and parses XML files
- mcpreadsourcecode: Reads and analyzes source code files
"""

import io
import json
import os
import sys
import tempfile
import traceback
from datetime import date, datetime
from typing import Any, Dict, List, Optional

import fitz
import html2text
import pandas as pd
import xmltodict
from bs4 import BeautifulSoup
from docx2markdown._docx_to_markdown import docx_to_markdown
from dotenv import load_dotenv
from mcp.server.fastmcp import FastMCP
from PIL import Image, ImageDraw, ImageFont
from pptx import Presentation
from pydantic import BaseModel, Field
from PyPDF2 import PdfReader
from tabulate import tabulate
from xls2xlsx import XLS2XLSX

from aworld.logs.util import logger
from aworld.utils import import_package
from mcp_servers.image_server import encode_images

mcp = FastMCP("document-server")


# Define model classes for different document types
class TextDocument(BaseModel):
    """Model representing a text document"""

    content: str
    file_path: str
    file_name: str
    file_size: int
    last_modified: str


class HtmlDocument(BaseModel):
    """Model representing an HTML document"""

    content: str  # Extracted text content
    html_content: str  # Original HTML content
    file_path: str
    file_name: str
    file_size: int
    last_modified: str
    title: Optional[str] = None
    links: Optional[List[Dict[str, str]]] = None
    images: Optional[List[Dict[str, str]]] = None
    tables: Optional[List[str]] = None
    markdown: Optional[str] = None  # HTML converted to Markdown format


class JsonDocument(BaseModel):
    """Model representing a JSON document"""

    format: str  # "json" or "jsonl"
    type: Optional[str] = None  # "array" or "object" for standard JSON
    count: Optional[int] = None
    keys: Optional[List[str]] = None
    data: Any
    file_path: str
    file_name: str


class XmlDocument(BaseModel):
    """Model representing an XML document"""

    content: Dict
    file_path: str
    file_name: str


class PdfImage(BaseModel):
    """Model representing an image extracted from a PDF"""

    page: int
    format: str
    width: int
    height: int
    path: str


class PdfDocument(BaseModel):
    """Model representing a PDF document"""

    content: str
    file_path: str
    file_name: str
    page_count: int
    images: Optional[List[PdfImage]] = None
    image_count: Optional[int] = None
    image_dir: Optional[str] = None
    error: Optional[str] = None


class PdfResult(BaseModel):
    """Model representing results from processing multiple PDF documents"""

    total_files: int
    success_count: int
    failed_count: int
    results: List[PdfDocument]


class DocxDocument(BaseModel):
    """Model representing a Word document"""

    content: str
    file_path: str
    file_name: str


class ExcelSheet(BaseModel):
    """Model representing a sheet in an Excel file"""

    name: str
    data: List[Dict[str, Any]]
    markdown_table: str
    row_count: int
    column_count: int


class ExcelDocument(BaseModel):
    """Model representing an Excel document"""

    file_name: str
    file_path: str
    processed_path: Optional[str] = None
    file_type: str
    sheet_count: int
    sheet_names: List[str]
    sheets: List[ExcelSheet]
    success: bool = True
    error: Optional[str] = None


class ExcelResult(BaseModel):
    """Model representing results from processing multiple Excel documents"""

    total_files: int
    success_count: int
    failed_count: int
    results: List[ExcelDocument]


class PowerPointSlide(BaseModel):
    """Model representing a slide in a PowerPoint presentation"""

    slide_number: int
    image: str  # Base64 encoded image


class PowerPointDocument(BaseModel):
    """Model representing a PowerPoint document"""

    file_path: str
    file_name: str
    slide_count: int
    slides: List[PowerPointSlide]


class SourceCodeDocument(BaseModel):
    """Model representing a source code document"""

    content: str
    file_type: str
    file_path: str
    file_name: str
    line_count: int
    size_bytes: int
    last_modified: str
    classes: Optional[List[str]] = None
    functions: Optional[List[str]] = None
    imports: Optional[List[str]] = None
    package: Optional[List[str]] = None
    methods: Optional[List[str]] = None
    includes: Optional[List[str]] = None


class DocumentError(BaseModel):
    """Model representing an error in document processing"""

    error: str
    file_path: Optional[str] = None
    file_name: Optional[str] = None


class ComplexEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, datetime):
            return o.strftime("%Y-%m-%d %H:%M:%S")
        elif isinstance(o, date):
            return o.strftime("%Y-%m-%d")
        else:
            return json.JSONEncoder.default(self, o)


def handle_error(e: Exception, error_type: str, file_path: Optional[str] = None) -> str:
    """Unified error handling and return standard format error message"""
    error_msg = f"{error_type} error: {str(e)}"
    logger.error(traceback.format_exc())

    error = DocumentError(
        error=error_msg,
        file_path=file_path,
        file_name=os.path.basename(file_path) if file_path else None,
    )

    return error.model_dump_json()


def check_file_readable(document_path: str) -> str:
    """Check if file exists and is readable, return error message or None"""
    if not os.path.exists(document_path):
        return f"File does not exist: {document_path}"
    if not os.access(document_path, os.R_OK):
        return f"File is not readable: {document_path}"
    return None


@mcp.tool(
    description="Read and return content from local text file. Cannot process https://URLs files."
)
def mcpreadtext(
    document_path: str = Field(description="The input local text file path."),
) -> str:
    """Read and return content from local text file. Cannot process https://URLs files."""
    error = check_file_readable(document_path)
    if error:
        return DocumentError(error=error, file_path=document_path).model_dump_json()

    try:
        with open(document_path, "r", encoding="utf-8") as f:
            content = f.read()

        result = TextDocument(
            content=content,
            file_path=document_path,
            file_name=os.path.basename(document_path),
            file_size=os.path.getsize(document_path),
            last_modified=datetime.fromtimestamp(
                os.path.getmtime(document_path)
            ).strftime("%Y-%m-%d %H:%M:%S"),
        )

        return result.model_dump_json()
    except Exception as e:
        return handle_error(e, "Text file reading", document_path)


@mcp.tool(
    description="Read and parse JSON or JSONL file, return the parsed content. Cannot process https://URLs files."
)
def mcpreadjson(
    document_path: str = Field(description="Local path to JSON or JSONL file"),
    is_jsonl: bool = Field(
        default=False,
        description="Whether the file is in JSONL format (one JSON object per line)",
    ),
) -> str:
    """Read and parse JSON or JSONL file, return the parsed content. Cannot process https://URLs files."""
    error = check_file_readable(document_path)
    if error:
        return DocumentError(error=error, file_path=document_path).model_dump_json()

    try:
        # Choose processing method based on file type
        if is_jsonl:
            # Process JSONL file (one JSON object per line)
            results = []
            with open(document_path, "r", encoding="utf-8") as f:
                for line_num, line in enumerate(f, 1):
                    line = line.strip()
                    if not line:
                        continue
                    try:
                        json_obj = json.loads(line)
                        results.append(json_obj)
                    except json.JSONDecodeError as e:
                        logger.warning(
                            f"JSON parsing error at line {line_num}: {str(e)}"
                        )

            # Create result model
            result = JsonDocument(
                format="jsonl",
                count=len(results),
                data=results,
                file_path=document_path,
                file_name=os.path.basename(document_path),
            )

        else:
            # Process standard JSON file
            with open(document_path, "r", encoding="utf-8") as f:
                data = json.load(f)

            # Create result model based on data type
            if isinstance(data, list):
                result = JsonDocument(
                    format="json",
                    type="array",
                    count=len(data),
                    data=data,
                    file_path=document_path,
                    file_name=os.path.basename(document_path),
                )
            else:
                result = JsonDocument(
                    format="json",
                    type="object",
                    keys=list(data.keys()) if isinstance(data, dict) else [],
                    data=data,
                    file_path=document_path,
                    file_name=os.path.basename(document_path),
                )

        return result.model_dump_json()

    except json.JSONDecodeError as e:
        return handle_error(e, "JSON parsing", document_path)
    except Exception as e:
        return handle_error(e, "JSON file reading", document_path)


@mcp.tool(
    description="Read and return content from XML file. return the parsed content. Cannot process https://URLs files."
)
def mcpreadxml(
    document_path: str = Field(description="The local input XML file path."),
) -> str:
    """Read and return content from XML file. Cannot process https://URLs files."""
    error = check_file_readable(document_path)
    if error:
        return DocumentError(error=error, file_path=document_path).model_dump_json()

    try:
        with open(document_path, "r", encoding="utf-8") as f:
            data = f.read()

        result = XmlDocument(
            content=xmltodict.parse(data),
            file_path=document_path,
            file_name=os.path.basename(document_path),
        )

        return result.model_dump_json()
    except Exception as e:
        return handle_error(e, "XML file reading", document_path)


@mcp.tool(
    description="Read and return content from PDF file with optional image extraction. return the parsed content. Cannot process https://URLs files."
)
def mcpreadpdf(
    document_paths: List[str] = Field(description="The local input PDF file paths."),
    extract_images: bool = Field(
        default=False, description="Whether to extract images from PDF (default: False)"
    ),
) -> str:
    """Read and return content from PDF file with optional image extraction. Cannot process https://URLs files."""
    try:

        results = []
        success_count = 0
        failed_count = 0

        for document_path in document_paths:
            error = check_file_readable(document_path)
            if error:
                results.append(
                    PdfDocument(
                        content="",
                        file_path=document_path,
                        file_name=os.path.basename(document_path),
                        page_count=0,
                        error=error,
                    )
                )
                failed_count += 1
                continue

            try:
                with open(document_path, "rb") as f:
                    reader = PdfReader(f)
                    content = " ".join(page.extract_text() for page in reader.pages)
                    page_count = len(reader.pages)

                    pdf_result = PdfDocument(
                        content=content,
                        file_path=document_path,
                        file_name=os.path.basename(document_path),
                        page_count=page_count,
                    )

                    # Extract images if requested
                    if extract_images:
                        images_data = []
                        # Use /tmp directory for storing images
                        output_dir = "/tmp/pdf_images"

                        # Create output directory if it doesn't exist
                        os.makedirs(output_dir, exist_ok=True)

                        # Generate a unique subfolder based on filename to avoid conflicts
                        pdf_name = os.path.splitext(os.path.basename(document_path))[0]
                        timestamp = datetime.now().strftime("%Y%m%d%H%M%S")
                        image_dir = os.path.join(output_dir, f"{pdf_name}_{timestamp}")
                        os.makedirs(image_dir, exist_ok=True)

                        try:
                            # Open PDF with PyMuPDF
                            pdf_document = fitz.open(document_path)

                            # Iterate through each page
                            for page_index in range(len(pdf_document)):
                                page = pdf_document[page_index]

                                # Get image list
                                image_list = page.get_images(full=True)

                                # Process each image
                                for img_index, img in enumerate(image_list):
                                    # Extract image information
                                    xref = img[0]
                                    base_image = pdf_document.extract_image(xref)
                                    image_bytes = base_image["image"]
                                    image_ext = base_image["ext"]

                                    # Save image to file in /tmp directory
                                    img_filename = f"pdf_image_p{page_index+1}_{img_index+1}.{image_ext}"
                                    img_path = os.path.join(image_dir, img_filename)

                                    with open(img_path, "wb") as img_file:
                                        img_file.write(image_bytes)
                                        logger.success(f"Image saved: {img_path}")

                                    # Get image dimensions
                                    with Image.open(img_path) as img:
                                        width, height = img.size

                                    # Add to results with file path instead of base64
                                    images_data.append(
                                        PdfImage(
                                            page=page_index + 1,
                                            format=image_ext,
                                            width=width,
                                            height=height,
                                            path=img_path,
                                        )
                                    )

                            pdf_result.images = images_data
                            pdf_result.image_count = len(images_data)
                            pdf_result.image_dir = image_dir

                        except Exception as img_error:
                            logger.error(f"Error extracting images: {str(img_error)}")
                            # Don't clean up on error so we can keep any successfully extracted images
                            pdf_result.error = str(img_error)

                results.append(pdf_result)
                success_count += 1

            except Exception as e:
                results.append(
                    PdfDocument(
                        content="",
                        file_path=document_path,
                        file_name=os.path.basename(document_path),
                        page_count=0,
                        error=str(e),
                    )
                )
                failed_count += 1

        # Create final result
        pdf_result = PdfResult(
            total_files=len(document_paths),
            success_count=success_count,
            failed_count=failed_count,
            results=results,
        )

        return pdf_result.model_dump_json()

    except Exception as e:
        return handle_error(e, "PDF file reading")


@mcp.tool(
    description="Read and return content from Word file. return the parsed content. Cannot process https://URLs files."
)
def mcpreaddocx(
    document_path: str = Field(description="The local input Word file path."),
) -> str:
    """Read and return content from Word file. Cannot process https://URLs files."""
    error = check_file_readable(document_path)
    if error:
        return DocumentError(error=error, file_path=document_path).model_dump_json()

    try:

        file_name = os.path.basename(document_path)
        md_file_path = f"{file_name}.md"
        docx_to_markdown(document_path, md_file_path)

        with open(md_file_path, "r", encoding="utf-8") as f:
            content = f.read()

        os.remove(md_file_path)

        result = DocxDocument(
            content=content, file_path=document_path, file_name=file_name
        )

        return result.model_dump_json()
    except Exception as e:
        return handle_error(e, "Word file reading", document_path)


@mcp.tool(
    description="Read multiple Excel/CSV files and convert sheets to Markdown tables. return the parsed content. Cannot process https://URLs files."
)
def mcpreadexcel(
    document_paths: List[str] = Field(
        description="List of local input Excel/CSV file paths."
    ),
    max_rows: int = Field(
        1000, description="Maximum number of rows to read per sheet (default: 1000)"
    ),
    convert_xls_to_xlsx: bool = Field(
        False,
        description="Whether to convert XLS files to XLSX format (default: False)",
    ),
) -> str:
    """Read multiple Excel/CSV files and convert sheets to Markdown tables. Cannot process https://URLs files."""
    try:

        # Import required packages
        import_package("tabulate")

        # Import xls2xlsx package if conversion is requested
        if convert_xls_to_xlsx:
            import_package("xls2xlsx")

        all_results = []
        temp_files = []  # Track temporary files for cleanup
        success_count = 0
        failed_count = 0

        # Process each file
        for document_path in document_paths:
            # Check if file exists and is readable
            error = check_file_readable(document_path)
            if error:
                all_results.append(
                    ExcelDocument(
                        file_name=os.path.basename(document_path),
                        file_path=document_path,
                        file_type="UNKNOWN",
                        sheet_count=0,
                        sheet_names=[],
                        sheets=[],
                        success=False,
                        error=error,
                    )
                )
                failed_count += 1
                continue

            try:
                # Check file extension
                file_ext = os.path.splitext(document_path)[1].lower()

                # Validate file type
                if file_ext not in [".csv", ".xls", ".xlsx", ".xlsm"]:
                    error_msg = f"Unsupported file format: {file_ext}. Only CSV, XLS, XLSX, and XLSM formats are supported."
                    all_results.append(
                        ExcelDocument(
                            file_name=os.path.basename(document_path),
                            file_path=document_path,
                            file_type=file_ext.replace(".", "").upper(),
                            sheet_count=0,
                            sheet_names=[],
                            sheets=[],
                            success=False,
                            error=error_msg,
                        )
                    )
                    failed_count += 1
                    continue

                # Convert XLS to XLSX if requested and file is XLS
                processed_path = document_path
                if convert_xls_to_xlsx and file_ext == ".xls":
                    try:
                        logger.info(f"Converting XLS to XLSX: {document_path}")
                        converter = XLS2XLSX(document_path)
                        # Create temp file with xlsx extension
                        xlsx_path = (
                            os.path.splitext(document_path)[0] + "_converted.xlsx"
                        )
                        converter.to_xlsx(xlsx_path)
                        processed_path = xlsx_path
                        temp_files.append(xlsx_path)  # Track for cleanup
                        logger.success(f"Converted XLS to XLSX: {xlsx_path}")
                    except Exception as conv_error:
                        logger.error(f"XLS to XLSX conversion error: {str(conv_error)}")
                        # Continue with original file if conversion fails

                excel_sheets = []
                sheet_names = []

                # Handle CSV files differently
                if file_ext == ".csv":
                    # For CSV files, create a single sheet with the file name
                    sheet_name = os.path.basename(document_path).replace(".csv", "")
                    df = pd.read_csv(processed_path, nrows=max_rows)

                    # Create markdown table
                    markdown_table = "*Empty table*"
                    if not df.empty:
                        headers = df.columns.tolist()
                        table_data = df.values.tolist()
                        markdown_table = tabulate(
                            table_data, headers=headers, tablefmt="pipe"
                        )

                        if len(df) >= max_rows:
                            markdown_table += (
                                f"\n\n*Note: Table truncated to {max_rows} rows*"
                            )

                    # Create sheet model
                    excel_sheets.append(
                        ExcelSheet(
                            name=sheet_name,
                            data=df.to_dict(orient="records"),
                            markdown_table=markdown_table,
                            row_count=len(df),
                            column_count=len(df.columns),
                        )
                    )

                    sheet_names = [sheet_name]

                else:
                    # For Excel files, process all sheets
                    with pd.ExcelFile(processed_path) as xls:
                        sheet_names = xls.sheet_names

                        for sheet_name in sheet_names:
                            # Read Excel sheet into DataFrame with row limit
                            df = pd.read_excel(
                                xls, sheet_name=sheet_name, nrows=max_rows
                            )

                            # Create markdown table
                            markdown_table = "*Empty table*"
                            if not df.empty:
                                headers = df.columns.tolist()
                                table_data = df.values.tolist()
                                markdown_table = tabulate(
                                    table_data, headers=headers, tablefmt="pipe"
                                )

                                if len(df) >= max_rows:
                                    markdown_table += f"\n\n*Note: Table truncated to {max_rows} rows*"

                            # Create sheet model
                            excel_sheets.append(
                                ExcelSheet(
                                    name=sheet_name,
                                    data=df.to_dict(orient="records"),
                                    markdown_table=markdown_table,
                                    row_count=len(df),
                                    column_count=len(df.columns),
                                )
                            )

                # Create result for this file
                file_result = ExcelDocument(
                    file_name=os.path.basename(document_path),
                    file_path=document_path,
                    processed_path=(
                        processed_path if processed_path != document_path else None
                    ),
                    file_type=file_ext.replace(".", "").upper(),
                    sheet_count=len(sheet_names),
                    sheet_names=sheet_names,
                    sheets=excel_sheets,
                    success=True,
                )

                all_results.append(file_result)
                success_count += 1

            except Exception as file_error:
                # Handle errors for individual files
                error_msg = str(file_error)
                logger.error(f"File reading error for {document_path}: {error_msg}")
                all_results.append(
                    ExcelDocument(
                        file_name=os.path.basename(document_path),
                        file_path=document_path,
                        file_type=os.path.splitext(document_path)[1]
                        .replace(".", "")
                        .upper(),
                        sheet_count=0,
                        sheet_names=[],
                        sheets=[],
                        success=False,
                        error=error_msg,
                    )
                )
                failed_count += 1

        # Clean up temporary files
        for temp_file in temp_files:
            try:
                if os.path.exists(temp_file):
                    os.remove(temp_file)
                    logger.info(f"Removed temporary file: {temp_file}")
            except Exception as cleanup_error:
                logger.warning(
                    f"Error cleaning up temporary file {temp_file}: {str(cleanup_error)}"
                )

        # Create final result
        excel_result = ExcelResult(
            total_files=len(document_paths),
            success_count=success_count,
            failed_count=failed_count,
            results=all_results,
        )

        return excel_result.model_dump_json()

    except Exception as e:
        return handle_error(e, "Excel/CSV files processing")


@mcp.tool(
    description="Read and convert PowerPoint slides to base64 encoded images. return the parsed content. Cannot process https://URLs files."
)
def mcpreadpptx(
    document_path: str = Field(description="The local input PowerPoint file path."),
) -> str:
    """Read and convert PowerPoint slides to base64 encoded images. Cannot process https://URLs files."""
    error = check_file_readable(document_path)
    if error:
        return DocumentError(error=error, file_path=document_path).model_dump_json()

    # Create temporary directory
    temp_dir = tempfile.mkdtemp()
    slides_data = []

    try:
        presentation = Presentation(document_path)
        total_slides = len(presentation.slides)

        if total_slides == 0:
            raise ValueError("PPT file does not contain any slides")

        # Process each slide
        for i, slide in enumerate(presentation.slides):
            # Set slide dimensions
            slide_width_px = 1920  # 16:9 ratio
            slide_height_px = 1080

            # Create blank image
            slide_img = Image.new("RGB", (slide_width_px, slide_height_px), "white")
            draw = ImageDraw.Draw(slide_img)
            font = ImageFont.load_default()

            # Draw slide number
            draw.text((20, 20), f"Slide {i+1}/{total_slides}", fill="black", font=font)

            # Process shapes in the slide
            for shape in slide.shapes:
                try:
                    # Process images
                    if hasattr(shape, "image") and shape.image:
                        image_stream = io.BytesIO(shape.image.blob)
                        img = Image.open(image_stream)
                        left = int(
                            shape.left * slide_width_px / presentation.slide_width
                        )
                        top = int(
                            shape.top * slide_height_px / presentation.slide_height
                        )
                        slide_img.paste(img, (left, top))

                    # Process text
                    elif hasattr(shape, "text") and shape.text:
                        text_left = int(
                            shape.left * slide_width_px / presentation.slide_width
                        )
                        text_top = int(
                            shape.top * slide_height_px / presentation.slide_height
                        )
                        draw.text(
                            (text_left, text_top),
                            shape.text,
                            fill="black",
                            font=font,
                        )

                except Exception as shape_error:
                    logger.warning(
                        f"Error processing shape in slide {i+1}: {str(shape_error)}"
                    )

            # Save slide image
            img_path = os.path.join(temp_dir, f"slide_{i+1}.jpg")
            slide_img.save(img_path, "JPEG")

            # Convert to base64
            base64_image = encode_images(img_path)
            slides_data.append(
                PowerPointSlide(
                    slide_number=i + 1, image=f"data:image/jpeg;base64,{base64_image}"
                )
            )

        # Create result
        result = PowerPointDocument(
            file_path=document_path,
            file_name=os.path.basename(document_path),
            slide_count=total_slides,
            slides=slides_data,
        )

        return result.model_dump_json()

    except Exception as e:
        return handle_error(e, "PowerPoint processing", document_path)
    finally:
        # Clean up temporary files
        try:
            for file in os.listdir(temp_dir):
                os.remove(os.path.join(temp_dir, file))
            os.rmdir(temp_dir)
        except Exception as cleanup_error:
            logger.warning(f"Error cleaning up temporary files: {str(cleanup_error)}")


@mcp.tool(
    description="Read HTML file and extract text content, optionally extract links, images, and table information, and convert to Markdown format."
)
def mcpreadhtmltext(
    document_path: str = Field(description="Local HTML file path or Web URL."),
    extract_links: bool = Field(
        default=True, description="Whether to extract link information"
    ),
    extract_images: bool = Field(
        default=True, description="Whether to extract image information"
    ),
    extract_tables: bool = Field(
        default=True, description="Whether to extract table information"
    ),
    convert_to_markdown: bool = Field(
        default=True, description="Whether to convert HTML to Markdown format"
    ),
) -> str:
    """Read HTML file and extract text content, optionally extract links, images, and table information, and convert to Markdown format."""
    error = check_file_readable(document_path)
    if error:
        return DocumentError(error=error, file_path=document_path).model_dump_json()

    try:

        # Read HTML file
        with open(document_path, "r", encoding="utf-8") as f:
            html_content = f.read()

        # Parse HTML using BeautifulSoup
        soup = BeautifulSoup(html_content, "html.parser")

        # Extract text content (remove script and style content)
        for script in soup(["script", "style"]):
            script.extract()
        text_content = soup.get_text(separator="\n", strip=True)

        # Extract title
        title = soup.title.string if soup.title else None

        # Initialize result object
        result = HtmlDocument(
            content=text_content,
            html_content=html_content,
            file_path=document_path,
            file_name=os.path.basename(document_path),
            file_size=os.path.getsize(document_path),
            last_modified=datetime.fromtimestamp(
                os.path.getmtime(document_path)
            ).strftime("%Y-%m-%d %H:%M:%S"),
            title=title,
        )

        # Extract links
        if extract_links:
            links = []
            for link in soup.find_all("a"):
                href = link.get("href")
                text = link.get_text(strip=True)
                if href:
                    links.append({"url": href, "text": text})
            result.links = links

        # Extract images
        if extract_images:
            images = []
            for img in soup.find_all("img"):
                src = img.get("src")
                alt = img.get("alt", "")
                if src:
                    images.append({"src": src, "alt": alt})
            result.images = images

        # Extract tables
        if extract_tables:
            tables = []
            for table in soup.find_all("table"):
                tables.append(str(table))
            result.tables = tables

        # Convert to Markdown
        if convert_to_markdown:
            h = html2text.HTML2Text()
            h.ignore_links = False
            h.ignore_images = False
            h.ignore_tables = False
            markdown_content = h.handle(html_content)
            result.markdown = markdown_content

        return result.model_dump_json()

    except Exception as e:
        return handle_error(e, "HTML file reading", document_path)


def main():
    load_dotenv()

    print("Starting Document MCP Server...", file=sys.stderr)
    mcp.run(transport="stdio")


# Make the module callable
def __call__():
    """
    Make the module callable for uvx.
    This function is called when the module is executed directly.
    """
    main()


sys.modules[__name__].__call__ = __call__

# Run the server when the script is executed directly
if __name__ == "__main__":
    main()



import asyncio
import logging
import abc
from typing import Dict, List, Any, Optional

from aworld.sandbox.base import Sandbox
from aworld.sandbox.models import SandboxStatus, SandboxEnvType, SandboxInfo
from aworld.sandbox.run.mcp_servers import McpServers


class BaseSandbox(Sandbox):
    """
    Base sandbox implementation with common functionality for all sandbox types.
    This class implements common methods and provides a foundation for specific sandbox implementations.
    """

    def __init__(
            self,
            sandbox_id: Optional[str] = None,
            env_type: Optional[int] = None,
            metadata: Optional[Dict[str, str]] = None,
            timeout: Optional[int] = None,
            mcp_servers: Optional[List[str]] = None,
            mcp_config: Optional[Any] = None,
    ):
        """
        Initialize a new BaseSandbox instance.
        
        Args:
            sandbox_id: Unique identifier for the sandbox. If None, one will be generated.
            env_type: The environment type (LOCAL, K8S, SUPERCOMPUTER).
            metadata: Additional metadata for the sandbox.
            timeout: Timeout for sandbox operations.
            mcp_servers: List of MCP servers to use.
            mcp_config: Configuration for MCP servers.
        """
        super().__init__(
            sandbox_id=sandbox_id,
            env_type=env_type,
            metadata=metadata,
            timeout=timeout,
            mcp_servers=mcp_servers,
            mcp_config=mcp_config
        )
        self._logger = self._setup_logger()
        
    def _setup_logger(self):
        """
        Set up a logger for the sandbox instance.
        
        Returns:
            logging.Logger: Configured logger instance.
        """
        logger = logging.getLogger(f"sandbox.{self.__class__.__name__}.{self.sandbox_id[:8]}")
        return logger
        
    def get_info(self) -> SandboxInfo:
        """
        Get information about the sandbox.
        
        Returns:
            SandboxInfo: Information about the sandbox.
        """
        return {
            "sandbox_id": self.sandbox_id,
            "status": self.status,
            "metadata": self.metadata,
            "env_type": self.env_type
        }
    
    @property
    def mcpservers(self) -> McpServers:
        """
        Module for running MCP servers in the sandbox.
        This property provides access to the MCP servers instance.
        
        Returns:
            McpServers: The MCP servers instance.
        """
        if hasattr(self, '_mcpservers'):
            return self._mcpservers
        return None
    
    @abc.abstractmethod
    async def cleanup(self) -> bool:
        """
        Clean up sandbox resources.
        This method must be implemented by subclasses to provide environment-specific cleanup.
        
        Returns:
            bool: True if cleanup was successful, False otherwise.
        """
        pass
    
    @abc.abstractmethod
    async def remove(self) -> bool:
        """
        Remove the sandbox.
        This method must be implemented by subclasses to provide environment-specific removal.
        
        Returns:
            bool: True if removal was successful, False otherwise.
        """
        pass 


import threading
from typing import Protocol, TypeVar, Any, Callable
from wrapt import wrap_function_wrapper
from concurrent import futures
import aworld.trace as trace
from aworld.trace.base import TraceContext, Span
from aworld.trace.propagator import get_global_trace_context
from aworld.trace.instrumentation import Instrumentor
from aworld.trace.instrumentation.utils import unwrap
from aworld.logs.util import logger


R = TypeVar("R")


class HasTraceContext(Protocol):
    _trace_context: TraceContext


class ThreadingInstrumentor(Instrumentor):
    '''
    Trace instrumentor for threading
    '''

    def instrumentation_dependencies(self) -> str:
        return ()

    def _instrument(self, **kwargs: Any):
        self._instrument_thread()
        self._instrument_timer()
        self._instrument_thread_pool()

    def _uninstrument(self, **kwargs: Any):
        self._uninstrument_thread()
        self._uninstrument_timer()
        self._uninstrument_thread_pool()

    @staticmethod
    def _instrument_thread():
        wrap_function_wrapper(
            threading.Thread,
            "start",
            ThreadingInstrumentor.__wrap_threading_start,
        )
        wrap_function_wrapper(
            threading.Thread,
            "run",
            ThreadingInstrumentor.__wrap_threading_run,
        )

    @staticmethod
    def _instrument_timer():
        wrap_function_wrapper(
            threading.Timer,
            "start",
            ThreadingInstrumentor.__wrap_threading_start,
        )
        wrap_function_wrapper(
            threading.Timer,
            "run",
            ThreadingInstrumentor.__wrap_threading_run,
        )

    @staticmethod
    def _instrument_thread_pool():
        wrap_function_wrapper(
            futures.ThreadPoolExecutor,
            "submit",
            ThreadingInstrumentor.__wrap_thread_pool_submit,
        )

    @staticmethod
    def _uninstrument_thread():
        unwrap(threading.Thread, "start")
        unwrap(threading.Thread, "run")

    @staticmethod
    def _uninstrument_timer():
        unwrap(threading.Timer, "start")
        unwrap(threading.Timer, "run")

    @staticmethod
    def _uninstrument_thread_pool():
        unwrap(futures.ThreadPoolExecutor, "submit")

    @staticmethod
    def __wrap_threading_start(
        call_wrapped: Callable[[], None],
        instance: HasTraceContext,
        args: tuple[()],
        kwargs: dict[str, Any],
    ) -> None:
        span: Span = trace.get_current_span()
        if span:
            instance._trace_context = TraceContext(
                trace_id=span.get_trace_id(), span_id=span.get_span_id())
        return call_wrapped(*args, **kwargs)

    @staticmethod
    def __wrap_threading_run(
        call_wrapped: Callable[..., R],
        instance: HasTraceContext,
        args: tuple[Any, ...],
        kwargs: dict[str, Any],
    ) -> R:

        token = None
        try:
            if hasattr(instance, "_trace_context"):
                if instance._trace_context:
                    token = get_global_trace_context().set(instance._trace_context)
            return call_wrapped(*args, **kwargs)
        finally:
            if token:
                get_global_trace_context().reset(token)

    @staticmethod
    def __wrap_thread_pool_submit(
        call_wrapped: Callable[..., R],
        instance: futures.ThreadPoolExecutor,
        args: tuple[Callable[..., Any], ...],
        kwargs: dict[str, Any],
    ) -> R:
        # obtain the original function and wrapped kwargs
        original_func = args[0]
        trace_context = None
        span: Span = trace.get_current_span()
        if span:
            trace_context = TraceContext(
                trace_id=span.get_trace_id(), span_id=span.get_span_id())

        def wrapped_func(*func_args: Any, **func_kwargs: Any) -> R:
            token = None
            try:
                if trace_context:
                    token = get_global_trace_context().set(trace_context)
                return original_func(*func_args, **func_kwargs)
            finally:
                if token:
                    get_global_trace_context().reset(token)

        # replace the original function with the wrapped function
        new_args: tuple[Callable[..., Any], ...] = (wrapped_func,) + args[1:]
        return call_wrapped(*new_args, **kwargs)


def instrument_theading(**kwargs: Any) -> None:
    ThreadingInstrumentor().instrument(**kwargs)
    logger.info("Threading instrumented")



import re
from typing import List
from aworld.trace.base import Propagator, Carrier, TraceContext
from aworld.trace.baggage import BaggageContext
from aworld.logs.util import logger
from urllib.parse import quote_plus, unquote_plus


class W3CBaggagePropagator(Propagator):
    """
    W3C baggage propagator.
    """

    _MAX_HEADER_LENGTH = 8192
    _MAX_PAIR_LENGTH = 4096
    _MAX_PAIRS = 180
    _BAGGAGE_HEADER_NAME = "baggage"
    _DELIMITER_PATTERN = re.compile(r"[ \t]*,[ \t]*")

    def extract(self, carrier: Carrier):
        """
        Extract the trace context from the carrier.
        Args:
            carrier: The carrier to extract the trace context from.
        """
        baggage_header = self._get_value(carrier, self._BAGGAGE_HEADER_NAME)
        if not baggage_header:
            return None

        if len(baggage_header) > self._MAX_HEADER_LENGTH:
            logger.warning(
                f"baggage header length exceeds {self._MAX_HEADER_LENGTH}")
            return None

        baggage_entries: List[str] = re.split(
            self._DELIMITER_PATTERN, baggage_header)
        if len(baggage_entries) > self._MAX_PAIRS:
            logger.warning(f"baggage entries exceeds {self._MAX_PAIRS}")

        for entry in baggage_entries:
            if len(entry) > self._MAX_PAIR_LENGTH:
                logger.warning(
                    f"baggage entry length exceeds {self._MAX_PAIR_LENGTH}")
                continue
            try:
                key, value = entry.split("=", 1)
                key = unquote_plus(key).strip()
                value = unquote_plus(value).strip()
            except ValueError:
                logger.warning(f"baggage entry format error: {entry}")
                continue
            BaggageContext.set_baggage(key, value)

    def inject(self, carrier: Carrier, context: TraceContext):
        """
        Inject the trace context into the carrier.
        Args:
            carrier: The carrier to inject the trace context into.
            context: The trace context to inject.
        """
        baggage = BaggageContext.get_baggage()
        if baggage:
            baggage_header = ",".join(
                f"{quote_plus(key)}={quote_plus(value)}" for key, value in baggage.items())
            carrier.set(self._BAGGAGE_HEADER_NAME, baggage_header)



# coding: utf-8
# Copyright (c) 2025 inclusionAI.

# need tool's name, desc and cls params
TOOL_TEMPLATE = """
import json
from typing import List, Tuple, Dict, Any, Union

from aworld.config.conf import ToolConfig, ConfigDict
from aworld.core.envs.tool import Tool, AsyncTool, ToolFactory
from aworld.core.common import Observation, ActionModel, ActionResult
from aworld.logs.util import logger
from aworld.virtual_environments.utils import build_observation
{action_import}


@ToolFactory.register(name="{name}", desc="{desc}", supported_action={action})
class {name}Tool({cls}[Observation, List[ActionModel]]):
    def __init__(self, conf: Union[Dict[str, Any], ConfigDict, ToolConfig], **kwargs) -> None:
        super().__init__(conf, **kwargs)

    {async_flag}def reset(self, *, seed: int | None = None, options: Dict[str, str] | None = None) -> Tuple[
        Observation, dict[str, Any]]:
        # from options obtain user query
        return build_observation(observer=self.name(),
                                 ability=''), dict()

    {async_flag}def step(self, action: List[ActionModel], **kwargs) -> Tuple[Observation, float, bool, bool, Dict[str, Any]]:
        reward = 0
        fail_error = ""
        action_result = None

        invalid_acts: List[int] = []
        for i, act in enumerate(action):
            if act.tool_name != "{name}":
                invalid_acts.append(i)

        if invalid_acts:
            for i in invalid_acts:
                action[i] = None

        resp = ""
        try:
            action_result, resp = {await_flag}self.action_executor.{async_underline}execute_action(action, **kwargs)
            reward = 1
        except Exception as e:
            fail_error = str(e)

        terminated = kwargs.get("terminated", False)
        if action_result:
            for res in action_result:
                if res.is_done:
                    terminated = res.is_done

        info = dict()
        info['exception'] = fail_error
        info.update(kwargs)
        if resp:
            resp = json.dumps(resp)
        else:
            resp = action_result[0].content

        action_result = [ActionResult(content=resp, keep=True, is_done=True)]
        observation = build_observation(observer=self.name(),
                                        ability=action[-1].action_name,
                                        content=resp,
                                        info=info)
        observation.action_result = action_result
        self._finished = True
        return (observation,
                reward,
                terminated,
                kwargs.get("truncated", False),
                dict())

    {async_flag}def close(self) -> None:
        pass

    {async_flag}def finished(self) -> bool:
        # one time
        return True
"""



import os
from re import compile as re_compile
from re import search
from typing import Final, Iterable, Any
from urllib.parse import urlparse, urlunparse, unquote
from wsgiref.types import WSGIEnvironment
from requests.models import PreparedRequest

HTTP_REQUEST_METHOD: Final = "http.request.method"
HTTP_FLAVOR: Final = "http.flavor"
HTTP_HOST: Final = "http.host"
HTTP_SCHEME: Final = "http.scheme"
HTTP_USER_AGENT: Final = "http.user_agent"
HTTP_SERVER_NAME: Final = "http.server_name"
SERVER_ADDRESS: Final = "server.address"
SERVER_PORT: Final = "server.port"
URL_PATH: Final = "url.path"
URL_QUERY: Final = "url.query"
CLIENT_ADDRESS: Final = "client.address"
CLIENT_PORT: Final = "client.port"
URL_FULL: Final = "url.full"

HTTP_REQUEST_BODY_SIZE: Final = "http.request.body.size"
HTTP_REQUEST_HEADER: Final = "http.request.header"
HTTP_REQUEST_SIZE: Final = "http.request.size"
HTTP_RESPONSE_BODY_SIZE: Final = "http.response.body.size"
HTTP_RESPONSE_HEADER: Final = "http.response.header"
HTTP_RESPONSE_SIZE: Final = "http.response.size"
HTTP_RESPONSE_STATUS_CODE: Final = "http.response.status_code"
HTTP_ROUTE = "http.route"


def collect_request_attributes(environ: WSGIEnvironment):

    attributes: dict[str] = {}

    request_method = environ.get("REQUEST_METHOD", "")
    request_method = request_method.upper()
    attributes[HTTP_REQUEST_METHOD] = request_method
    attributes[HTTP_FLAVOR] = environ.get("SERVER_PROTOCOL", "")
    attributes[HTTP_SCHEME] = environ.get("wsgi.url_scheme", "")
    attributes[HTTP_SERVER_NAME] = environ.get("SERVER_NAME", "")
    attributes[HTTP_HOST] = environ.get("HTTP_HOST", "")
    host_port = environ.get("SERVER_PORT")
    if host_port:
        attributes[SERVER_PORT] = host_port
    target = environ.get("RAW_URI")
    if target is None:
        target = environ.get("REQUEST_URI")
    if target:
        path, query = _parse_url_query(target)
        attributes[URL_PATH] = path
        attributes[URL_QUERY] = query
    remote_addr = environ.get("REMOTE_ADDR", "")
    attributes[CLIENT_ADDRESS] = remote_addr
    attributes[CLIENT_PORT] = environ.get("REMOTE_PORT", "")
    remote_host = environ.get("REMOTE_HOST")
    if remote_host and remote_host != remote_addr:
        attributes[CLIENT_ADDRESS] = remote_host
    attributes[HTTP_USER_AGENT] = environ.get("HTTP_USER_AGENT", "")
    return attributes


def collect_attributes_from_request(request: PreparedRequest) -> dict[str]:
    attributes: dict[str] = {}

    url = remove_url_credentials(request.url)
    attributes[HTTP_REQUEST_METHOD] = request.method
    attributes[URL_FULL] = url
    parsed_url = urlparse(url)
    if parsed_url.scheme:
        attributes[HTTP_SCHEME] = parsed_url.scheme
    if parsed_url.hostname:
        attributes[HTTP_HOST] = parsed_url.hostname
    if parsed_url.port:
        attributes[SERVER_PORT] = parsed_url.port
    return attributes


def url_disabled(url: str, excluded_urls: Iterable[str]) -> bool:
    """
    Check if the url is disabled.
    Args:
        url: The url to check.
        excluded_urls: The excluded urls.
    Returns:
        True if the url is disabled, False otherwise.
    """
    if excluded_urls is None:
        return False
    regex = re_compile("|".join(excluded_urls))
    return search(regex, url)


def get_excluded_urls(instrumentation: str) -> list[str]:
    """
    Get the excluded urls.
    Args:
        instrumentation: The instrumentation to get the excluded urls for.
    Returns:
        The excluded urls.
    """

    excluded_urls = os.environ.get(f"{instrumentation}_EXCLUDED_URLS")

    return parse_excluded_urls(excluded_urls)


def parse_excluded_urls(excluded_urls: str) -> list[str]:
    """
    Parse the excluded urls.
    Args:
        excluded_urls: The excluded urls.
    Returns:
        The excluded urls.
    """
    if excluded_urls:
        excluded_url_list = [
            excluded_url.strip() for excluded_url in excluded_urls.split(",")
        ]
    else:
        excluded_url_list = []

    return excluded_url_list


def remove_url_credentials(url: str) -> str:
    """Given a string url, remove the username and password only if it is a valid url"""

    try:
        parsed = urlparse(url)
        if all([parsed.scheme, parsed.netloc]):  # checks for valid url
            parsed_url = urlparse(url)
            _, _, netloc = parsed.netloc.rpartition("@")
            return urlunparse(
                (
                    parsed_url.scheme,
                    netloc,
                    parsed_url.path,
                    parsed_url.params,
                    parsed_url.query,
                    parsed_url.fragment,
                )
            )
    except ValueError:  # an unparsable url was passed
        pass
    return url


def parser_host_port_url_from_asgi(scope: dict[str, Any]):
    """Returns (host, port, full_url) tuple."""
    server = scope.get("server") or ["0.0.0.0", 80]
    port = server[1]
    server_host = server[0] + (":" + str(port) if str(port) != "80" else "")
    full_path = scope.get("path", "")
    http_url = scope.get("scheme", "http") + "://" + server_host + full_path
    return server_host, port, http_url


def collect_request_attributes_asgi(scope: dict[str, Any]):
    attributes: dict[str] = {}
    server_host, port, http_url = parser_host_port_url_from_asgi(scope)
    query_string = scope.get("query_string")
    if query_string and http_url:
        if isinstance(query_string, bytes):
            query_string = query_string.decode("utf8")
        http_url += "?" + unquote(query_string)
    attributes[HTTP_REQUEST_METHOD] = scope.get("method", "")
    attributes[HTTP_FLAVOR] = scope.get("http_version", "")
    attributes[HTTP_SCHEME] = scope.get("scheme", "")
    attributes[HTTP_HOST] = server_host
    attributes[SERVER_PORT] = port
    attributes[URL_FULL] = remove_url_credentials(http_url)
    attributes[URL_PATH] = scope.get("path", "")
    header = scope.get("headers")
    if header:
        for key, value in header:
            if key == b"user-agent":
                attributes[HTTP_USER_AGENT] = value.decode("utf8")

    client = scope.get("client")
    if client:
        attributes[CLIENT_ADDRESS] = client[0]
        attributes[CLIENT_PORT] = client[1]

    return attributes


def _parse_url_query(url: str):
    parsed_url = urlparse(url)
    path = parsed_url.path
    query_params = parsed_url.query
    return path, query_params



# coding: utf-8
# Copyright (c) 2025 inclusionAI.
from typing import Union, Dict, AsyncGenerator, Any

from aworld.core.agent.base import is_agent
from aworld.core.common import ActionModel, TaskItem
from aworld.core.context.base import Context
from aworld.core.event.base import Message, Constants
from aworld.core.tool.base import AsyncTool, Tool, ToolFactory
from aworld.logs.util import logger
from aworld.runners.handler.base import DefaultHandler
from aworld.runners.utils import TaskType


class DefaultToolHandler(DefaultHandler):
    def __init__(self, tools: Dict[str, Union[Tool, AsyncTool]], tools_conf: Dict[str, Any]):
        self.tools = tools
        self.tools_conf = tools_conf

    async def handle(self, message: Message) -> AsyncGenerator[Message, None]:
        if message.category != Constants.TOOL:
            return

        # data is List[ActionModel]
        data = message.payload
        if not data:
            # error message, p2p
            yield Message(
                category=Constants.TASK,
                payload=TaskItem(msg="no data to process.", data=data, stop=True),
                sender='agent_handler',
                session_id=Context.instance().session_id,
                topic=TaskType.ERROR
            )
            return

        for action in data:
            if not isinstance(action, ActionModel):
                # error message, p2p
                yield Message(
                    category=Constants.TASK,
                    payload=TaskItem(msg="action not a ActionModel.", data=data, stop=True),
                    sender=self.name(),
                    session_id=Context.instance().session_id,
                    topic=TaskType.ERROR
                )
                return

        new_tools = dict()
        tool_mapping = dict()
        # Directly use or use tools after creation.
        for act in data:
            if is_agent(act):
                logger.warning(f"somethings wrong, {act} is an agent.")
                continue

            if not self.tools or (self.tools and act.tool_name not in self.tools):
                # dynamic only use default config in module.
                conf = self.tools_conf.get(act.tool_name)
                tool = ToolFactory(act.tool_name, conf=conf, asyn=conf.use_async if conf else False)
                tool.event_driven = True
                if isinstance(tool, Tool):
                    tool.reset()
                elif isinstance(tool, AsyncTool):
                    await tool.reset()
                tool_mapping[act.tool_name] = []
                self.tools[act.tool_name] = tool
                new_tools[act.tool_name] = tool
            if act.tool_name not in tool_mapping:
                tool_mapping[act.tool_name] = []
            tool_mapping[act.tool_name].append(act)

        if new_tools:
            yield Message(
                category=Constants.TASK,
                payload=TaskItem(data=new_tools),
                sender=self.name(),
                session_id=Context.instance().session_id,
                topic=TaskType.SUBSCRIBE_TOOL
            )

        for tool_name, actions in tool_mapping.items():
            if not (isinstance(self.tools[tool_name], Tool) or isinstance(self.tools[tool_name], AsyncTool)):
                logger.warning(f"Unsupported tool type: {self.tools[tool_name]}")
                continue

            # send to the tool
            yield Message(
                category=Constants.TOOL,
                payload=actions,
                sender=actions[0].agent_name if actions else '',
                session_id=Context.instance().session_id,
                receiver=tool_name
            )



from typing import Sequence
from pydantic import BaseModel, Field, root_validator
from typing import Optional


class MetricTemplate(BaseModel):
    """
    MetricTemplate is a class for defining a metric template.
    """
    type: str
    name: str
    description: Optional[str] = None
    unit: Optional[str] = Field(default="1")
    labels: Optional[list[str]] = None
    buckets: Optional[Sequence[float]] = None

    @root_validator(pre=True)
    def set_default_description(cls, values):
        """
        Set the default description if it is not set.
        """
        if 'description' not in values or values['description'] is None:
            values['description'] = values['name']
        return values


class MetricTemplates:
    REQUEST_COUNT = MetricTemplate(**{
        "type": "counter",
        "name": "request_count",
        "description": "The number of requests received",
        "unit": "1",
        "labels": ["method", "status"]
    })

    REQUEST_LATENCY = MetricTemplate(**{
        "type": "histogram",
        "name": "request_latency",
        "description": "The latency of requests",
        "unit": "ms",
        "labels": ["method", "status"],
        # "buckets": [0.01, 0.05, 0.1, 0.5, 1, 5, 10, 50, 100, 500, 1000]
    })



import abc
from collections import Counter
from typing import (
    Any,
    List,
    Dict,
    Generator,
    AsyncGenerator,
)

from aworld.models.model_response import ModelResponse


class LLMProviderBase(abc.ABC):
    """Base class for large language model providers, defines unified interface."""

    def __init__(self,
                 api_key: str = None,
                 base_url: str = None,
                 model_name: str = None,
                 sync_enabled: bool = None,
                 async_enabled: bool = None,
                 **kwargs):
        """Initialize provider.

        Args:
            api_key: API key.
            base_url: Service URL.
            model_name: Model name.
            **kwargs: Other parameters.
        """
        self.api_key = api_key
        self.base_url = base_url
        self.model_name = model_name
        self.kwargs = kwargs
        # Determine whether to initialize sync and async providers
        self.need_sync = sync_enabled if sync_enabled is not None else async_enabled is not True
        self.need_async = async_enabled if async_enabled is not None else sync_enabled is not True

        # Initialize providers based on flags
        self.provider = self._init_provider() if self.need_sync else None
        self.async_provider = self._init_async_provider() if self.need_async else None
        self.stream_tool_buffer=[]

    @abc.abstractmethod
    def _init_provider(self):
        """Initialize specific provider instance, to be implemented by subclasses.
        Returns:
            Provider instance.
        """

    def _init_async_provider(self):
        """Initialize async provider instance. Optional for subclasses that don't need async support.
        Only called if async provider is needed.

        Returns:
            Async provider instance.
        """
        return None

    @classmethod
    def supported_models(cls) -> list[str]:
        return []

    def preprocess_messages(self, messages: List[Dict[str, str]]) -> Any:
        """Preprocess messages, convert OpenAI format messages to specific provider required format.

        Args:
            messages: OpenAI format message list [{"role": "system", "content": "..."}, ...].

        Returns:
            Converted messages, format determined by specific provider.
        """
        return messages

    @abc.abstractmethod
    def postprocess_response(self, response: Any) -> ModelResponse:
        """Post-process response, convert provider response to unified ModelResponse.

        Args:
            response: Original response from provider.

        Returns:
            ModelResponse: Unified format response object.
            
        Raises:
            LLMResponseError: When LLM response error occurs.
        """

    def postprocess_stream_response(self, chunk: Any) -> ModelResponse:
        """Post-process streaming response chunk, convert provider chunk to unified ModelResponse.

        Args:
            chunk: Original response chunk from provider.
            
        Returns:
            ModelResponse: Unified format response object for the chunk.
            
        Raises:
            LLMResponseError: When LLM response error occurs.
        """

    async def acompletion(self,
                          messages: List[Dict[str, str]],
                          temperature: float = 0.0,
                          max_tokens: int = None,
                          stop: List[str] = None,
                          **kwargs) -> ModelResponse:
        """Asynchronously call model to generate response.
        
        Args:
            messages: Message list, format is [{"role": "system", "content": "..."}, {"role": "user", "content": "..."}].
            temperature: Temperature parameter.
            max_tokens: Maximum number of tokens to generate.
            stop: List of stop sequences.
            **kwargs: Other parameters.

        Returns:
            ModelResponse: Unified model response object.
            
        Raises:
            LLMResponseError: When LLM response error occurs.
            RuntimeError: When async provider is not initialized.
        """
        if not self.async_provider:
            raise RuntimeError(
                "Async provider not initialized. Make sure 'async_enabled' parameter is set to True in initialization.")


    @abc.abstractmethod
    def completion(self,
                   messages: List[Dict[str, str]],
                   temperature: float = 0.0,
                   max_tokens: int = None,
                   stop: List[str] = None,
                   **kwargs) -> ModelResponse:
        """Synchronously call model to generate response.

        Args:
            messages: Message list, format is [{"role": "system", "content": "..."}, {"role": "user", "content": "..."}].
            temperature: Temperature parameter.
            max_tokens: Maximum number of tokens to generate.
            stop: List of stop sequences.
            **kwargs: Other parameters.

        Returns:
            ModelResponse: Unified model response object.
            
        Raises:
            LLMResponseError: When LLM response error occurs.
            RuntimeError: When sync provider is not initialized.
        """

    def stream_completion(self,
                          messages: List[Dict[str, str]],
                          temperature: float = 0.0,
                          max_tokens: int = None,
                          stop: List[str] = None,
                          **kwargs) -> Generator[ModelResponse, None, None]:
        """Synchronously call model to generate streaming response.

        Args:
            messages: Message list, format is [{"role": "system", "content": "..."}, {"role": "user", "content": "..."}].
            temperature: Temperature parameter.
            max_tokens: Maximum number of tokens to generate.
            stop: List of stop sequences.
            **kwargs: Other parameters.

        Returns:
            Generator yielding ModelResponse chunks.
            
        Raises:
            LLMResponseError: When LLM response error occurs.
            RuntimeError: When sync provider is not initialized.
        """

    async def astream_completion(self,
                                 messages: List[Dict[str, str]],
                                 temperature: float = 0.0,
                                 max_tokens: int = None,
                                 stop: List[str] = None,
                                 **kwargs) -> AsyncGenerator[ModelResponse, None]:
        """Asynchronously call model to generate streaming response.

        Args:
            messages: Message list, format is [{"role": "system", "content": "..."}, {"role": "user", "content": "..."}].
            temperature: Temperature parameter.
            max_tokens: Maximum number of tokens to generate.
            stop: List of stop sequences.
            **kwargs: Other parameters.

        Returns:
            AsyncGenerator yielding ModelResponse chunks.
            
        Raises:
            LLMResponseError: When LLM response error occurs.
            RuntimeError: When async provider is not initialized.
        """
        return None

    def _accumulate_chunk_usage(self, usage: Dict[str, int], chunk_usage: Dict[str, int]):
        """Accumulate usage statistics from chunk into the main usage dictionary.

        Args:
            usage: Dictionary to accumulate usage into (will be modified)
            chunk_usage: Usage statistics from the current chunk
        """
        if not chunk_usage:
            return

        usage.update(dict(Counter(usage) + Counter(chunk_usage)))



# coding: utf-8
# Copyright (c) 2025 inclusionAI.
import json
from typing import List, Tuple, Dict, Any

from aworld.core.tool.base import Tool
from aworld.core.common import Observation, ActionModel, Config
from aworld.logs.util import logger
from aworld.tools.utils import build_observation


class TemplateTool(Tool):
    def __init__(self, conf: Config, **kwargs) -> None:
        super(TemplateTool, self).__init__(conf, **kwargs)

    def reset(self, *, seed: int | None = None, options: Dict[str, str] | None = None) -> Tuple[
        Observation, dict[str, Any]]:
        # from options obtain user query
        return build_observation(observer=self.name(),
                                 ability='',
                                 content=options.get("query", None) if options else None), {}

    def do_step(self,
                action: List[ActionModel],
                **kwargs) -> Tuple[Observation, float, bool, bool, Dict[str, Any]]:
        reward = 0
        fail_error = ""
        action_result = None

        invalid_acts: List[int] = []
        for i, act in enumerate(action):
            if act.tool_name != self.name():
                logger.warning(f"tool {act.tool_name} is not a {self.name()} tool!")
                invalid_acts.append(i)

        if invalid_acts:
            for i in invalid_acts:
                action[i] = None

        resp = ""
        try:
            action_result, resp = self.action_executor.execute_action(action, **kwargs)
            reward = 1
        except Exception as e:
            fail_error = str(e)

        terminated = kwargs.get("terminated", False)
        for res in action_result:
            if res.is_done:
                terminated = res.is_done
                self._finished = True

        info = {"exception": fail_error}
        info.update(kwargs)
        if resp:
            resp = json.dumps(resp)
        else:
            resp = action_result[0].content

        observation = build_observation(observer=self.name(),
                                        action_result=action_result,
                                        ability=action[-1].action_name,
                                        content=resp)
        return (observation,
                reward,
                terminated,
                kwargs.get("truncated", False),
                info)

    def close(self) -> None:
        pass

    def finished(self) -> bool:
        # one time
        return True



# coding: utf-8
# Copyright (c) 2025 inclusionAI.
from collections import Counter, OrderedDict
from typing import Dict

from aworld.config import ConfigDict
from aworld.core.context.session import Session
from aworld.core.singleton import InheritanceSingleton
from aworld.utils.common import nest_dict_counter


class Context(InheritanceSingleton):
    def __init__(self, user: str = None, **kwargs):
        self._user = user
        self._init(**kwargs)

    def _init(self, **kwargs):
        self._task_id = kwargs.get('task_id')
        self._engine = kwargs.get('engine')
        self._trace_id = kwargs.get('trace_id')
        self._session: Session = kwargs.get('session')
        self.context_info = ConfigDict()
        self.agent_info = ConfigDict()
        self.trajectories = OrderedDict()
        self._token_usage = {
            "completion_tokens": 0,
            "prompt_tokens": 0,
            "total_tokens": 0
        }

    def add_token(self, usage: Dict[str, int]):
        self._token_usage = nest_dict_counter(self._token_usage, usage)

    def reset(self, **kwargs):
        self._init(**kwargs)

    @property
    def trace_id(self):
        return self._trace_id

    @trace_id.setter
    def trace_id(self, trace_id):
        self._trace_id = trace_id

    @property
    def token_usage(self):
        return self._token_usage

    @property
    def engine(self):
        return self._engine

    @property
    def user(self):
        return self._user

    @user.setter
    def user(self, user):
        if user is not None:
            self._user = user

    @property
    def task_id(self):
        return self._task_id

    @task_id.setter
    def task_id(self, task_id):
        if task_id is not None:
            self._task_id = task_id

    @property
    def session_id(self):
        if self.session:
            return self.session.session_id
        else:
            return None

    @property
    def session(self):
        return self._session

    @session.setter
    def session(self, session: Session):
        self._session = session

    @property
    def record_path(self):
        return "."

    @property
    def is_task(self):
        return True

    @property
    def enable_visible(self):
        return False

    @property
    def enable_failover(self):
        return False

    @property
    def enable_cluster(self):
        return False




import os
from urllib.parse import urljoin
from typing import Optional, Sequence
from typing_extensions import LiteralString
from uuid import uuid4
from opentelemetry import metrics
from opentelemetry.sdk.resources import Resource
from opentelemetry.semconv.resource import ResourceAttributes
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.metrics.export import PeriodicExportingMetricReader, ConsoleMetricExporter
from aworld.metrics.metric import (
    Gauge,
    Histogram,
    MetricProvider,
    Counter,
    MetricExporter,
    UpDownCounter,
    get_metric_provider,
    set_metric_provider
)

MEMORY_FIELDS: list[LiteralString] = 'available used free active inactive buffers cached shared wired slab'.split()
"""
The fields of the memory information returned by psutil.virtual_memory().
"""


class OpentelemetryMetricProvider(MetricProvider):
    """
    MetricProvider is a class for providing metrics.
    """

    def __init__(self, exporter: MetricExporter = None):
        """Initialize the MetricProvider.
        Args:
            exporter: The exporter of the metric.
        """
        super().__init__()
        if not exporter:
            exporter = ConsoleMetricExporter()
        self._exporter = exporter

        self._otel_provider = MeterProvider(
            metric_readers=[PeriodicExportingMetricReader(
                exporter=self._exporter, export_interval_millis=5000)],
            resource=build_otel_resource()
        )
        metrics.set_meter_provider(self._otel_provider)
        self._meter = self._otel_provider.get_meter("aworld")

    def create_counter(self,
                       name: str,
                       description: str,
                       unit: str,
                       labelnames: Optional[Sequence[str]] = None) -> Counter:
        """
        Create a counter.
        Args:
            name: The name of the counter.
            description: The description of the counter.
            unit: The unit of the counter.
        """
        return OpentelemetryCounter(name, description, unit, self)

    def create_un_down_counter(self,
                               name: str,
                               description: str,
                               unit: str,
                               labelnames: Optional[Sequence[str]] = None) -> UpDownCounter:
        """
        Create a un-down counter.
        Args:
            name: The name of the counter.
            description: The description of the counter.
            unit: The unit of the counter.
        """
        return OpentelemetryUpDownCounter(name, description, unit, self)

    def create_gauge(self,
                     name: str,
                     description: str,
                     unit: str,
                     labelnames: Optional[Sequence[str]] = None) -> Gauge:
        """
        Create a gauge.
        Args:
            name: The name of the gauge.
            description: The description of the gauge.
            unit: The unit of the gauge.
        """
        return OpentelemetryGauge(name, description, unit, self)

    def create_histogram(self,
                         name: str,
                         description: str,
                         unit: str,
                         buckets: Optional[Sequence[float]] = None,
                         labelnames: Optional[Sequence[str]] = None) -> Histogram:
        """
        Create a histogram.
        Args:
            name: The name of the histogram.
            description: The description of the histogram.
            unit: The unit of the histogram.
            buckets: The buckets of the histogram.
        """
        return OpentelemetryHistogram(name, description, unit, self, buckets)

    def shutdown(self):
        """
        Shutdown the metric provider.
        """
        self._exporter.shutdown()
        self._otel_provider.shutdown()


class OpentelemetryCounter(Counter):
    """
    OpentelemetryCounter is a subclass of Counter, representing a counter metric.
    A counter is a cumulative metric that represents a single numerical value that only ever goes up.
    """

    def __init__(self,
                 name: str,
                 description: str,
                 unit: str,
                 provider: OpentelemetryMetricProvider):
        """
        Initialize the Counter.
        Args:
            name: The name of the counter.
            description: The description of the counter.
            unit: The unit of the counter.
            provider: The provider of the counter.
        """
        super().__init__(name, description, unit, provider)
        self._counter = provider._meter.create_counter(
            name=name, description=description, unit=unit)

    def add(self, value: int, labels: dict = None) -> None:
        """
        Add a value to the counter.
        Args:
            value: The value to add to the counter.
            labels: The labels to associate with the value.
        """
        if labels is None:
            labels = {}
        self._counter.add(value, labels)


class OpentelemetryUpDownCounter(UpDownCounter):
    """
    OpentelemetryUpDownCounter is a subclass of UpDownCounter, representing an un-down counter metric.
    An un-down counter is a cumulative metric that represents a single numerical value that only ever goes up.
    """

    def __init__(self,
                 name: str,
                 description: str,
                 unit: str,
                 provider: OpentelemetryMetricProvider):
        """
        Initialize the UnDownCounter. 
        Args:
            name: The name of the counter.
            description: The description of the counter.
            unit: The unit of the counter.
            provider: The provider of the counter.
        """
        super().__init__(name, description, unit, provider)
        self._counter = provider._meter.create_up_down_counter(
            name=name, description=description, unit=unit)

    def inc(self, value: int, labels: dict = None) -> None:
        """
        Add a value to the counter.
        Args:
            value: The value to add to the counter.
            labels: The labels to associate with the value.
        """
        if labels is None:
            labels = {}
        self._counter.add(value, labels)

    def dec(self, value: int, labels: dict = None) -> None:
        """
        Subtract a value from the counter.
        Args:
            value: The value to subtract from the counter.
            labels: The labels to associate with the value.
        """
        if labels is None:
            labels = {}
        self._counter.add(-value, labels)


class OpentelemetryGauge(Gauge):
    """
    OpentelemetryGauge is a subclass of Gauge, representing a gauge metric.
    A gauge is a metric that represents a single numerical value that can arbitrarily go up and down.
    """

    def __init__(self,
                 name: str,
                 description: str,
                 unit: str,
                 provider: OpentelemetryMetricProvider):
        """
        Initialize the Gauge.
        Args:
            name: The name of the gauge.
            description: The description of the gauge.
            unit: The unit of the gauge.
            provider: The provider of the gauge.
        """
        super().__init__(name, description, unit, provider)
        self._gauge = provider._meter.create_gauge(
            name=name, description=description, unit=unit)

    def set(self, value: int, labels: dict = None) -> None:
        """
        Set the value of the gauge.
        Args:
            value: The value to set the gauge to.
            labels: The labels to associate with the value.
        """
        if labels is None:
            labels = {}
        self._gauge.set(value, labels)


class OpentelemetryHistogram(Histogram):
    """
    OpentelemetryHistogram is a subclass of Histogram, representing a histogram metric.
    A histogram is a metric that represents the distribution of a set of values.     
    """

    def __init__(self,
                 name: str,
                 description: str,
                 unit: str,
                 provider: OpentelemetryMetricProvider,
                 buckets: Sequence[float] = None):
        """
        Initialize the Histogram.
        Args:
            name: The name of the histogram.
            description: The description of the histogram.
            unit: The unit of the histogram.
            provider: The provider of the histogram.
            buckets: The buckets of the histogram.
        """
        super().__init__(name, description, unit, provider, buckets)
        self._histogram = provider._meter.create_histogram(name=name,
                                                           description=description,
                                                           unit=unit,
                                                           explicit_bucket_boundaries_advisory=buckets)

    def record(self, value: int, labels: dict = None) -> None:
        """
        Record a value in the histogram.
        Args:
            value: The value to record in the histogram.
            labels: The labels to associate with the value.
        """
        if labels is None:
            labels = {}
        self._histogram.record(value, labels)


def configure_otlp_provider(backend: Sequence[str] = None,
                            base_url: str = None,
                            write_token: str = None,
                            **kwargs
                            ) -> None:
    """
    Configure the OpenTelemetry provider.
    Args:
        backends: The backends to use.
        base_url: The base URL of the backend.
        write_token: The write token of the backend.
        **kwargs: The keyword arguments to pass to the backend.
    """
    import requests
    from opentelemetry.exporter.otlp.proto.http import Compression
    from opentelemetry.exporter.otlp.proto.http.metric_exporter import OTLPMetricExporter

    if backend == "console":
        set_metric_provider(OpentelemetryMetricProvider())
    elif backend == "logfire":
        base_url = base_url or "https://logfire-us.pydantic.dev"
        headers = {'User-Agent': f'logfire/3.14.0',
                   'Authorization': write_token}
        session = requests.Session()
        session.headers.update(headers)
        exporter = OTLPMetricExporter(
            endpoint=urljoin(base_url, '/v1/metrics'),
            session=session,
            compression=Compression.Gzip,
        )
        set_metric_provider(OpentelemetryMetricProvider(exporter))
    elif backend == "antmonitor":
        ant_otlp_endpoint = os.getenv("ANT_OTEL_ENDPOINT")
        base_url = base_url or ant_otlp_endpoint
        session = requests.Session()
        session.timeout = 30
        exporter = OTLPMetricExporter(
            endpoint=base_url,
            session=session,
            compression=Compression.Gzip,
            timeout=30
        )
        set_metric_provider(OpentelemetryMetricProvider(exporter))

    if os.getenv("METRICS_SYSTEM_ENABLED") == "true":
        instrument_system_metrics()


def instrument_system_metrics():
    """
    Instrument system metrics.
    """
    try:
        from opentelemetry.instrumentation.system_metrics import (
            _DEFAULT_CONFIG,
            SystemMetricsInstrumentor
        )
    except ImportError:
        raise ImportError(
            "Could not import opentelemetry.instrumentation.system_metrics, please install it with `pip install opentelemetry-instrumentation-system-metrics`"
        )
    config = _DEFAULT_CONFIG.copy()
    config['system.memory.usage'] = MEMORY_FIELDS + ['total']
    config['system.memory.utilization'] = MEMORY_FIELDS
    config['system.swap.utilization'] = ['used']
    instrumentor = SystemMetricsInstrumentor(config=config)
    otel_provider = get_metric_provider()._otel_provider
    instrumentor.instrument(meter_provider=otel_provider)


def build_otel_resource():
    """
    Build the OpenTelemetry resource.
    """
    service_name = os.getenv("MONITOR_SERVICE_NAME") or "aworld"
    return Resource(
        attributes={
            ResourceAttributes.SERVICE_NAME: service_name,
            ResourceAttributes.SERVICE_NAMESPACE: "aworld",
            ResourceAttributes.SERVICE_INSTANCE_ID: uuid4().hex
        }
    )



import json
import logging
from builtins import anext
from datetime import datetime
from typing import Any, Dict, Generator, AsyncGenerator, Optional

from pydantic import Field, BaseModel, model_validator

from aworld.models.model_response import ModelResponse, ToolCall


class OutputPart(BaseModel):
    content: Any
    metadata: Optional[Dict[str, Any]] = Field(default_factory=dict, description="metadata")

    @model_validator(mode='after')
    def setup_metadata(self):
        # Ensure metadata is initialized
        if self.metadata is None:
            self.metadata = {}
        return self
    

class Output(BaseModel):
    metadata: Optional[Dict[str, Any]] = Field(default_factory=dict, description="metadata")
    parts: Any = Field(default_factory=list, exclude=True, description="parts of Output")
    data: Any = Field(default=None, exclude=True, description="Output Data")

    @model_validator(mode='after')
    def setup_defaults(self):
        # Ensure metadata and parts are initialized
        if self.metadata is None:
            self.metadata = {}
        if self.parts is None:
            self.parts = []
        return self

    def add_part(self, content: Any):
        if self.parts is None:
            self.parts = []
        self.parts.append(OutputPart(content=content))

    def output_type(self):
        return "default"


class ToolCallOutput(Output):

    @classmethod
    def from_tool_call(cls, tool_call: ToolCall):
        return cls(data = tool_call)

    def output_type(self):
        return "tool_call"

class ToolResultOutput(Output):

    origin_tool_call: ToolCall

    image: str = Field(default=None)

    images: list = Field(default=list)

    tool_type: str = Field(default=None)

    tool_name: str = Field(default=None)

    def output_type(self):
        return "tool_call_result"

    pass

class RunFinishedSignal(Output):

    def output_type(self):
        return "finished_signal"

RUN_FINISHED_SIGNAL = RunFinishedSignal()


class MessageOutput(Output):

    """
    MessageOutput structure of LLM output
    if you want to get the only response, you must first call reasoning_generator or set parameter only_response to True , then call response_generator
    if you model not reasoning, you do not need care about reasoning_generator and reasoning

    1. source: async/sync generator of the message
    2. reasoning_generator: async/sync reasoning generator of the message
    3. response_generator: async/sync response generator of the message;
    4. reasoning: reasoning of the message
    5. response: response of the message
    6. tool_calls

    """

    source: Any = Field(default=None, exclude=True, description="Source of the message")
    
    reason_generator: Any = Field(default=None, exclude=True, description="reasoning generator of the message")
    response_generator: Any = Field(default=None, exclude=True, description="response generator of the message")

    """
    result
    """
    reasoning: str = Field(default=None, description="reasoning of the message")
    response: Any = Field(default=None, description="response of the message")
    tool_calls: list[ToolCallOutput] = Field(default_factory=list, description="tool_calls")


    """
    other config
    """
    reasoning_format_start: str = Field(default="<think>", description="reasoning format start of the message")
    reasoning_format_end: str = Field(default="</think>", description="reasoning format end of the message")

    json_parse: bool = Field(default=False, description="json parse of the message", exclude=True)
    has_reasoning: bool = Field(default=False, description="has reasoning of the message")
    finished: bool = Field(default=False, description="finished of the message")

    @model_validator(mode='after')
    def setup_generators(self):
        """
        Setup generators for reasoning and response
        """
        source = self.source

        # if ModelResponse
        if isinstance(self.source, ModelResponse):
            source = self.source.content
            if self.source.tool_calls:
                [self.tool_calls.append(ToolCallOutput.from_tool_call(tool_call)) for tool_call in
                 self.source.tool_calls]

        if source is not None and isinstance(source, AsyncGenerator):
            # Create empty generators first, they will be initialized when actually used
            self.reason_generator = self.__aget_reasoning_generator()
            self.response_generator = self.__aget_response_generator()
        elif source is not None and isinstance(source, Generator):
            self.reason_generator, self.response_generator = self.__split_reasoning_and_response__()
        elif source is not None and isinstance(source, str):
            self.reasoning, self.response = self.__resolve_think__(source)
        return self

    async def get_finished_reasoning(self):
        if self.reasoning:
            return self.reasoning
        else:
            if self.has_reasoning and not self.reasoning:
                async for reason in self.reason_generator:
                    pass
                return self.reasoning
            else:
                return self.reasoning

    async def get_finished_response(self):
        if self.response:
            return self.response
        else:
            if self.response_generator:
                async for item in self.response_generator:
                    pass
            return self.response
    
    async def __aget_reasoning_generator(self) -> AsyncGenerator[str, None]:
        """
        Get reasoning content as async generator
        """
        if not self.has_reasoning:
            yield ""
            self.reasoning = ""
            return  
        
        reasoning_buffer = ""
        is_in_reasoning = False
        if self.reasoning and len(self.reasoning) > 0:
            yield self.reasoning
            return
        
        try:
            while True:
                chunk = await anext(self.source)
                chunk_content = self.get_chunk_content(chunk)
                if not chunk_content:
                    continue
                if chunk_content.startswith(self.reasoning_format_start):
                    is_in_reasoning = True
                    reasoning_buffer = chunk_content
                    yield chunk_content
                elif chunk_content.endswith(self.reasoning_format_end) and is_in_reasoning:
                    reasoning_buffer += chunk_content
                    yield chunk_content
                    self.reasoning = reasoning_buffer
                    break
                elif is_in_reasoning:
                    reasoning_buffer += chunk_content
                    yield chunk_content
        except StopAsyncIteration:
            logging.info("StopAsyncIteration")

    async def __aget_response_generator(self) -> AsyncGenerator[str, None]:
        """
        Get response content as async generator

        if has_reasoning is True, system will first call reasoning_generator if you not call it;
        else it will return content contains reasoning and response
        """
        response_buffer = ""

        if self.response and len(self.response) > 0:
            yield self.response
            return
        
        # if has_reasoning is True, system will first call reasoning_generator if you not call it;
        if self.has_reasoning and not self.reasoning:
            async for reason in self.reason_generator:
                pass

        try:
            while True:
                chunk = await anext(self.source)
                chunk_content = self.get_chunk_content(chunk)

                if not chunk_content:
                    continue
                response_buffer += chunk_content
                yield chunk_content
        except StopAsyncIteration:
            self.finished = True
            self.response = self.__resolve_json__(response_buffer, self.json_parse)

    def get_chunk_content(self, chunk):
        if isinstance(chunk, ModelResponse):
            return chunk.content
        else:
            return chunk

    def __split_reasoning_and_response__(self) -> tuple[Generator[str, None, None], Generator[str, None, None]]: # type: ignore
        """
        Split source into reasoning and response generators for sync source
        Returns:
            tuple: (reasoning_generator, response_generator)
        """
        if not self.has_reasoning:
            yield ""
            self.reasoning = ""
            return  
        
        if not isinstance(self.source, Generator):
            raise ValueError("Source must be a Generator")

        def reasoning_generator():
            if self.reasoning and len(self.reasoning) > 0:
                yield self.reasoning
                return

            reasoning_buffer = ""
            is_in_reasoning = False
            
            try:
                while True:
                    chunk = next(self.source)
                    chunk_content = self.get_chunk_content(chunk)
                    if chunk_content.startswith(self.reasoning_format_start):
                        is_in_reasoning = True
                        reasoning_buffer = chunk_content
                        yield chunk_content
                    elif chunk_content.endswith(self.reasoning_format_end) and is_in_reasoning:
                        reasoning_buffer += chunk_content
                        self.reasoning = reasoning_buffer
                        yield chunk_content
                        break
                    elif is_in_reasoning:
                        yield chunk_content
                        reasoning_buffer += chunk_content
            except StopIteration:
                print("StopIteration")
                self.reasoning = reasoning_buffer

        def response_generator():
            if self.response and len(self.response) > 0:
                yield self.response
                return
            
            # if has_reasoning is True, system will first call reasoning_generator if you not call it;
            if self.has_reasoning and not self.reasoning:
                for reason in self.reason_generator:
                    pass
            
            response_buffer = ""
            try:
                while True:
                    chunk = next(self.source)
                    chunk_content = self.get_chunk_content(chunk)
                    response_buffer += chunk_content
                    self.response = response_buffer
                    yield chunk_content
            except StopIteration:
                self.response = self.__resolve_json__(response_buffer,self.json_parse)
                self.finished = True


        return reasoning_generator(), response_generator()

    def __resolve_think__(self, content):
        import re
        start_tag = self.reasoning_format_start.replace("<", "").replace(">", "")
        end_tag = self.reasoning_format_end.replace("<", "").replace(">", "")

        llm_think = ""
        match = re.search(
            rf"<{re.escape(start_tag)}(.*?)>(.|\n)*?<{re.escape(end_tag)}>",
            content,
            flags=re.DOTALL,
        )
        if match:
            llm_think = match.group(0).replace("<think>", "").replace("</think>", "")
        llm_result = re.sub(
            rf"<{re.escape(start_tag)}(.*?)>(.|\n)*?<{re.escape(end_tag)}>",
            "",
            content,
            flags=re.DOTALL,
        )
        llm_result = self.__resolve_json__(llm_result, self.json_parse)

        return llm_think, llm_result

    def __resolve_json__(self, content, json_parse = False):
        if json_parse:
            if content.__contains__("```json"):
                content = content.replace("```json", "").replace("```", "")
            return json.loads(content)
        return content

    def output_type(self):
        return "message_output"

class StepOutput(Output):
    name: str
    step_num: int
    status: Optional[str] = Field(default="START", description="step_status")
    started_at: str = Field(default_factory=lambda: datetime.now().isoformat(), description="started at")
    finished_at: str = Field(default_factory=lambda: datetime.now().isoformat(), description="finished at")

    @classmethod
    def build_start_output(cls, name, step_num, data = None):
        return cls(name=name,step_num=step_num, data=data)

    @classmethod
    def build_finished_output(cls, name, step_num, data = None):
        return cls(name=name, step_num=step_num, status='FINISHED', data=data)

    @classmethod
    def build_failed_output(cls, name, step_num, data = None):
        return cls(name=name,step_num=step_num, status='FAILED', data=data)

    def output_type(self):
        return "step_output"



class SearchItem(BaseModel):
    title: str = Field(default="", description="search result title")
    url: str = Field(default="", description="search result url")
    content: str = Field(default="", description="search content", exclude=True)
    raw_content: Optional[str] = Field(default="", description="search raw content", exclude=True)
    metadata: Optional[Dict[str, Any]] = Field(default_factory=dict, description="metadata")


class SearchOutput(ToolResultOutput):
    query: str = Field(..., description="Search query string")
    results: list[SearchItem] = Field(default_factory=list, description="List of search results")

    @classmethod
    def from_dict(cls, data: dict) -> "SearchOutput":
        if not isinstance(data, dict):
            data = {}

        query = data.get("query")
        if query is None:
            raise ValueError("query is required")

        results_data = data.get("results", [])
        
        search_items = []
        for result in results_data:
            if isinstance(result, SearchItem):
                search_items.append(result)
            elif isinstance(result, dict):
                search_items.append(SearchItem(**result))
            else:
                raise ValueError(f"Invalid result type: {type(result)}")

        return cls(
            query=query,
            results=search_items
        )

    def output_type(self):
        return "search_output"


import wrapt
from typing import Any, Collection
from aworld.trace.instrumentation import Instrumentor
from aworld.trace.base import Tracer, get_tracer_provider_silent
from aworld.trace.propagator import get_global_trace_propagator, get_global_trace_context
from aworld.trace.propagator.carrier import DictCarrier
from aworld.events.manager import EventManager
from aworld.core.event.base import Message
from aworld.logs.util import logger


def _emit_message_class_wrapper(tracer: Tracer):
    async def awrapper(wrapped, instance, args, kwargs):
        try:
            event = kwargs.get("event")
            propagator = get_global_trace_propagator()
            if propagator and event and isinstance(event, Message):
                if not event.headers:
                    event.headers = {}
                propagator.inject(DictCarrier(event.headers))
        except Exception as e:
            logger.error(
                f"EventManager emit_message trace propagate exception: {e}")
        return await wrapped(*args, **kwargs)
    return awrapper


def _emit_message_instance_wrapper(tracer: Tracer):

    @wrapt.decorator
    async def awrapper(wrapped, instance, args, kwargs):
        wrapper = _emit_message_class_wrapper(tracer)
        return await wrapper(wrapped, instance, args, kwargs)

    return awrapper


def _consume_class_wrapper(tracer: Tracer):
    async def awrapper(wrapped, instance, args, kwargs):
        event = await wrapped(*args, **kwargs)
        try:
            propagator = get_global_trace_propagator()
            if propagator and event and isinstance(event, Message) and event.headers:
                trace_context = propagator.extract(DictCarrier(event.headers))
                logger.info(
                    f"extract trace_context from event: {trace_context}")
                if trace_context:
                    get_global_trace_context().set(trace_context)
        except Exception as e:
            logger.error(
                f"EventManager consume trace propagate exception: {e}")
        return event
    return awrapper


def _consume_instance_wrapper(tracer: Tracer):

    @wrapt.decorator
    async def awrapper(wrapped, instance, args, kwargs):
        wrapper = _consume_class_wrapper(tracer)
        return await wrapper(wrapped, instance, args, kwargs)

    return awrapper


class EventBusInstrumentor(Instrumentor):

    def instrumentation_dependencies(self) -> Collection[str]:
        return ()

    def _uninstrument(self, **kwargs: Any):
        pass

    def _instrument(self, **kwargs: Any):
        tracer_provider = get_tracer_provider_silent()
        if not tracer_provider:
            return
        tracer = tracer_provider.get_tracer(
            "aworld.trace.instrumentation.eventbus")

        wrapt.wrap_function_wrapper(
            "aworld.events.manager",
            "EventManager.emit_message",
            _emit_message_class_wrapper(tracer=tracer)
        )

        wrapt.wrap_function_wrapper(
            "aworld.events.manager",
            "EventManager.consume",
            _consume_class_wrapper(tracer=tracer)
        )


def wrap_event_manager(manager: EventManager):
    tracer_provider = get_tracer_provider_silent()
    if not tracer_provider:
        return manager
    tracer = tracer_provider.get_tracer(
        "aworld.trace.instrumentation.eventbus")

    emit_wrapper = _emit_message_instance_wrapper(tracer)
    consume_wrapper = _consume_instance_wrapper(tracer)

    manager.emit_message = emit_wrapper(manager.emit_message)
    manager.consume = consume_wrapper(manager.consume)

    return manager
